from .setup import Indices, Models, Rho
from .propagate import TimeEvolve
from .util import basis, create, destroy, qeye, tensor, sigmam, sigmap, sigmaz
from multiprocessing import set_start_method
set_start_method('fork')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Standard example for calculating time evolution with pibs code.

"""

import numpy as np
import os
from pibs import Indices, Rho, Models
from pibs import TimeEvolve
from pibs import qeye, create, destroy, sigmam, sigmap, tensor, basis
import sys
import matplotlib.pyplot as plt
import scipy.sparse as sp

if __name__ == '__main__':
    # Set parameters
    ntls = 5                    # Number of two-level systems (TLS)
    nphot =ntls+1               # Photon space truncation
    w0 = 0.35                   # Level splitting of TLS
    wc = 0.0                    # Cavity frequency
    Omega =0.4                  # Vacuum Rabi splitting
    g = Omega / np.sqrt(ntls)   # Light-matter coupling 
    kappa = 0.01                # Cavity loss
    gamma = 0.001               # Molecular loss
    gamma_phi=0.0075            # Dephasing
    
    dt = 0.2        # timestep
    tmax = 200
    chunksize=200   # time chunks for parallel evolution
    
    # Solver parameters
    atol=1e-12
    rtol=1e-12
    nsteps=1000
    
    # Setup indices mapping
    indi = Indices(ntls,nphot, debug=False, save =True)
    
    # rotation matrix around x-axis of spin 1/2 : exp(-i*theta*Sx)=exp(-i*theta/2*sigmax) = cos(theta/2)-i*sin(theta/2)*sigmax
    theta = 0.25*np.pi
    rot_x = np.array([[np.cos(theta/2), -1j*np.sin(theta/2)],[-1j*np.sin(theta/2), np.cos(theta/2)]])
    rot_x_dag = np.array([[np.cos(theta/2), 1j*np.sin(theta/2)],[1j*np.sin(theta/2), np.cos(theta/2)]])
    
    rho_phot = basis(nphot,0) # Initial photon state with zero photons
    rho_spin = sp.csr_matrix(rot_x @ basis(2,0) @ rot_x_dag) # Initial spin state. 0 = up, 1 = down
    
    rho = Rho(rho_phot, rho_spin, indi, max_nrs=1) # Setup density matrix
    
    # Setup Liouvillian
    L = Models(wc, w0,g, kappa, gamma_phi,gamma,indi, parallel=0,progress=False, debug=False,save=True, num_cpus=None)
    L.setup_L_Tavis_Cummings(progress=False)
    
    # Operators for time evolution
    adag = create(nphot)
    a = destroy(nphot)
    n = adag*a
    p = tensor(qeye(nphot), sigmap()*sigmam())
    ops = [n,p] # operators to calculate expectations for
    
    # Time evolution
    evolve = TimeEvolve(rho, L, tmax, dt, atol=atol, rtol=rtol, nsteps=nsteps)
    #evolve.time_evolve_block_interp(ops, progress = True, expect_per_nu=False, start_block=None, save_states=False)
    evolve.time_evolve_chunk_parallel(ops, chunksize=chunksize, progress=True, num_cpus=None)
    
    # Get expectation values
    e_phot_tot = evolve.result.expect[0].real
    e_excit_site = evolve.result.expect[1].real
    t = evolve.result.t
    
    
    # Plot
    fig, ax = plt.subplots(1,2, figsize=(8,3))
    ax[0].plot(t, e_phot_tot/ntls)
    ax[1].plot(t, e_excit_site)
    ax[0].set_xlabel('t')
    ax[1].set_xlabel('t')
    ax[0].set_ylabel(r'$\langle n\rangle /N$')
    ax[1].set_ylabel(r'$\langle\sigma^+\sigma^-\rangle$')
    plt.tight_layout()
    plt.show()

    







#!/usr/bin/env python
from time import time
import pickle
import numpy as np
import scipy.sparse as sp
from math import factorial

def get_size(obj):
    """Estimate size of object in MB"""
    return round(len(pickle.dumps(obj))/1024**2, 2)
def export(obj, fp):
    pass

def timeit(func, args=None, msg=''):
    print(f'Running {msg} with arguments {args}...', flush=True)
    t0 = time()
    func(*args)
    elapsed = time()-t0
    print(f'Complete {elapsed:.0f}s', flush=True)

# other helper functions
def degeneracy_outer_invariant_optimized(outer1, outer2, inner):
    """ calculate how many distinct permutations there are of the spins (outer1, inner)
    and (inner, outer2), which leave outer1 and outer2 invariant. Necessary
    for commutator term in Liouvillian -i[H,rho]"""
    xi = outer1+ 2*outer2
    deg = 1
    for i in range(4):
        l = np.where(xi==i)[0]
        # print('loop',i)
        # print(l)
        if len(l) == 0:
            continue
        
        sub_inner = inner[l]
        s = sum(sub_inner)
        factor =  factorial(len(l)) / (factorial(s)*factorial(len(l)-s))
        deg = deg * factor
        
    return int(deg)


def states_compatible(state1, state2):
    """ checks, if state1 and state2 are equivalent up to permutation of spins"""
    if state1[0] != state2[0]:
        return False
    
    from numpy import where, setdiff1d, intersect1d
    
    spins1 = state1[1:]
    spins2 = state2[1:]
    if(sum(state1) != sum(state2)):
        return False
    
    return True

def permute_compatible(comp1, comp2, permute):
    """comp1 and comp2 contain compatible (=equal up to permutation) spin states.
    Find the permutation, that transforms comp2 in comp1 and perform the same
    transformation to permute. This is important to calculate the proper H-element
    in calc_L_line_block1."""
    
    # check indices, where the arrays have ones
    idx1 = np.where(comp1 == 1)[0]
    idx2 = np.where(comp2 == 1)[0]
    
    # find common indices and remove them, because they are already in order
    common_elements = np.intersect1d(idx1,idx2)
    idx_ones1 = np.setdiff1d(idx1, common_elements)
    idx_ones2 = np.setdiff1d(idx2, common_elements)
    
    cp_permute = np.copy(permute)
    for i in range(len(idx_ones1)):
        cp_permute[idx_ones2[i]] = permute[idx_ones1[i]]
        cp_permute[idx_ones1[i]] = permute[idx_ones2[i]]
        
    return cp_permute

def degeneracy_spin_gamma(spin1, spin2):
    """Return number of indices where spin1 and spin2 are up. For L0 part of
    individual loss L[sigmam]"""
    return np.count_nonzero(spin1+spin2==0)

def degeneracy_gamma_changing_block_efficient(outer1, outer2, inner1, inner2):
    """Find simultaneous permutation of inner1 and inner2, such that all but one
    spin index align, and in exactly the same positions. This is necessary
    for calculating the Lindblad operator of sigma minus (individual decay). efficient way """
    from itertools import permutations
    from numpy import where
    Oc = outer1 + 2*outer2
    Ic = inner1 + 2*inner2
    
    outer_num3 = len(where(Oc==3)[0])
    inner_num3 = len(where(Ic==3)[0])
    
    if outer_num3 - inner_num3 == 1:
        return outer_num3
    else:
        return 0


# operators here?

def qeye(N):

    return sp.eye(N, N, dtype=complex, format='csr')

def destroy(N):

    return sp.spdiags(np.sqrt(range(0, N)),
                           1, N, N, format='csr')

def create(N):

    qo = destroy(N)  # create operator using destroy function
    qo = qo.T.tocsr()  # transpose data in Qobj and convert to csr
    return qo

def sigmap():
    return sp.spdiags(np.array([0.0,1.0]),
                      1, 2, 2, format='csr')
                      
def sigmam():
    return sigmap().T.tocsr()

def sigmaz():
    return sp.spdiags(np.array([1.0,-1.0]),
                      0, 2, 2, format='csr')
def sigmax():
    return sigmap() + sigmam()

def sigmay():
    return -1j*sigmap() + 1j*sigmam()
    
def tensor(*args):

    if not args:
        raise TypeError("Requires at least one input argument")

    if len(args) == 1 and isinstance(args[0], (list, np.ndarray)):
        # this is the case when tensor is called on the form:
        # tensor([q1, q2, q3, ...])
        qlist = args[0]

    # elif len(args) == 1 and isinstance(args[0], Qobj):
    #     # tensor is called with a single input, do nothing
    #     return args[0]

    else:
        # this is the case when tensor is called on the form:
        # tensor(q1, q2, q3, ...)
        qlist = args

    out = []

    for n, q in enumerate(qlist):
        if n == 0:
            out = q
        else:
            out = sp.kron(out, q, format='csr')

    return out
    
def basis(N, n=0):
    """Create Fock density matrix for N-level Hilbert space
    with excitation in level n""" 
    
    if (not isinstance(N, (int, np.integer))) or N < 0:
        raise ValueError("N must be integer N >= 0")

    if (not isinstance(n, (int, np.integer))):
        raise ValueError("n must be integer n >= 0")

    if n > (N - 1):  # check if n is within bounds
        raise ValueError("basis vector index need to be in n <= N-1")

    bas = sp.lil_matrix((N, 1))  # column vector of zeros
    bas[n, 0] = 1  # 1 located at position n
    bas = bas.tocsr()

    return tensor(bas, bas.T)

def thermal_dm(N, nth):
    """ Create thermal density matreix for N-level Hilbert space with average
    photon number nth"""
    if (not isinstance(N, (int, np.integer))) or N < 0:
        raise ValueError("N must be integer N >= 0")

    dm = np.zeros((N,N))
    x= nth /(1+nth)
    Z = (1 - x**N) / (1 - x)  # partition function
    for i in range(N):
       dm[i,i] = 1.0 / Z * x**i
        
    return sp.csr_matrix(dm)
        
        
    
def expect(oper, state):

    # calculates expectation value via TR(op*rho)
    
    return ((oper@state).toarray()).trace()
    
    # return (oper.dot(state).toarray()).trace()

def vector_to_operator(op):

    n = int(np.sqrt(op.shape[0]))
    q = sp_reshape(op.T, (n, n)).T  # This seems to be quite expensive operation 
    return q


def sp_reshape(A, shape, format='csr'):

    if not hasattr(shape, '__len__') or len(shape) != 2:
        raise ValueError('Shape must be a list of two integers')

    C = sp.coo_matrix(A)
    nrows, ncols = C.shape
    size = nrows * ncols
    new_size = shape[0] * shape[1]

    if new_size != size:
        raise ValueError('Total size of new array must be unchanged.')

    flat_indices = ncols * C.row + C.col
    new_row, new_col = divmod(flat_indices, shape[1])
    B = sp.coo_matrix((C.data, (new_row, new_col)), shape=shape)

    if format == 'csr':
        return B.tocsr()
    elif format == 'coo':
        return B
    elif format == 'csc':
        return B.tocsc()
    elif format == 'lil':
        return B.tolil()
    else:
        raise ValueError('Return format not valid.')

def _multinominal(bins):
    """calculate multinominal coeffcient"""
    
    from math import factorial
    
    n = sum(bins)
    combinations  = factorial(n)
    for count_bin in range(len(bins)):
        combinations = combinations//factorial(bins[count_bin])
    return combinations


def wigner_d(theta, j, n, m):
    """ Calculate small Wigner d matrix d_{n,m}^j(theta) """
    from math import factorial
    smin = int(max(0, m-n))
    smax = int(min(j+m, j-n))
    print(smin,smax)

    
    #prefactor
    p = np.sqrt( float(factorial(int(j+n)))) * np.sqrt(float(factorial(int(j-n)))) * np.sqrt(float(factorial(int(j+m)))) * np.sqrt(float(factorial(int(j-m)) ))
    
    # sum
    d = 0
    for s in range(smin, smax+1):
        pre = factorial(int(j+m-s))*factorial(int(s))*factorial(int(n-m+s))*factorial(int(j-n-s))
        d += (-1)**(n-m+s) * np.cos(theta/2)**(2*j+m-n-2*s)*np.sin(theta/2)**(n-m+2*s) / pre
    
    d = d * p
    return d
    


class Progress:
    def __init__(self, total, description='', start_step=0):
        self.description = description
        self.step = start_step
        self.end = total-1
        self.percent = self.calc_percent()
        self.started = False

    def calc_percent(self):
        return int(100*self.step/self.end)

    def update(self, step=None):
        # print a description at the start of the calculation
        if not self.started:
            print('{}{:4d}%'.format(self.description, self.percent), end='', flush=True)
            self.started = True
            return
        # progress one step or to the specified step
        if step is None:
            self.step += 1
        else:
            self.step = step
        percent = self.calc_percent()
        # only waste time printing if % has actually increased one integer
        if percent > self.percent:
            print('\b\b\b\b\b{:4d}%'.format(percent), end='', flush=True)
            self.percent = percent
        if self.step == self.end:
            print('', flush=True)



#!/usr/bin/env python
import numpy as np
from time import time
from .util import  basis, vector_to_operator, expect, Progress
#from pibs.util import tensor, qeye, destroy, create, sigmap, sigmam, basis, sigmaz, vector_to_operator, expect

from scipy.integrate import ode
from scipy.interpolate import interp1d

from multiprocessing import Pool

 
class Results:
    def __init__(self):
        self.rho= []
        self.t = []
        self.expect = []
        self.expect_per_nu = []
        
class TimeEvolve():
    """ Class to handle time evolution of the quantum system. There are two
    main methods to solve the time evolution:
        
    time_evolve_block_interp:
        This method solves the block structure block for block sequentially.
        Meaning, starting with the block of highest excitation number nu_max, we solve
        the whole time evolution 
        
                d/dt rho_nu_max = L0_nu_max * rho_nu_max
                
        All states rho_nu_max are stored. Then we can iteratively solve for the lower
        blocks using
        
                d/dt rho_nu = L0_nu * rho_nu + L1_{nu+1} * rho_{nu+1}
        
        
    time_evolve_chunk_parallel2:
        This method sovles the block structure in a parallel manner. The time
        evolution is segmented into chunks of given number of time steps. In 
        the first time chunk, we solve the system for rho_nu_max. In the second
        time chunk, we solve the system for rho_nu_max, and additionally we can
        solve for rho_nu_max-1 for the first chunk, by using the result of rho_nu_max.
        That way, we maximize the amount of calculations we can do in parallel.
        Further, we can compute the expectation values for every chunk and then 
        discard the states, such that memory usage can be minimized.
        """
        
    def __init__(self, rho, L, tend, dt, atol=1e-5, rtol=1e-5, nsteps=500):
        """
        Initialize

        Parameters
        ----------
        rho : setup.Rho
            Density matrix object
        L : setup.Models
            Liouvillian object
        tend : float
            Final time for time evolution
        dt : float
            Time step
        atol : float, optional
            Absolute tolerance for integrator. The default is 1e-5.
        rtol : float, optional
            Relative tolerance for integrator. The default is 1e-5.
        nsteps : int, optional
            Max iterations for integrator. The default is 500.

        Returns
        -------
        None.

        """
        self.tend = tend
        self.dt = dt
        self.atol = atol
        self.rtol = rtol
        
        self.result = Results()
        
        # should we store rho and L like this ? Or give as parameters when calling .time_evolve_block?
        self.rho = rho # this makes just a reference to the rho object.
        self.L = L
        
        # fix this later
        self.indices = rho.indices
        
        self.nsteps=nsteps # number of steps per solver iteration. Standard value is 500 see https://stackoverflow.com/questions/40788747/internal-working-of-scipy-integrate-ode
        
        
        
    @staticmethod
    def evolve_nu_parallel2(args_tuple):
        """
        Handles the time evolution of block nu_max and all other blocks nu, too.
        Distinguish between those two cases by checking,if a feedforward input is given.
        
        This combined function enables the use of the multiprocessing Pool class.

        """
        tstart = time()
        rho0, rhoff_func, t0, nu, chunksize, store_initial, rtol, atol, method = args_tuple
        dt = t[1]-t[0]
        
        if rhoff_func is None: # no feedforward -> block nu_max
            nu_max = nu
            blocksize= len(mapping_block[nu_max])
            
            # initialize rho and times
            rho = np.zeros((blocksize, chunksize), dtype = complex)
            solver_times = np.zeros(chunksize)
        
            # integrator
            r = ode(_intfunc).set_integrator('zvode', method = method, atol=atol, rtol=rtol)
            r.set_initial_value(rho0,t0).set_f_params(L.L0[nu_max])
            
            # integrate
            if store_initial:
                rho[:,0] = rho0
                solver_times[0] = t0
                n_t = 1
            else:
                n_t=0
            while r.successful() and n_t < chunksize:
                rho_step = r.integrate(r.t+dt)
                rho[:,n_t] = rho_step
                solver_times[n_t] = r.t
                n_t += 1
            
            # elapsed = time() -tstart
            # print(f'Block numax={nu}: elapsed {elapsed:.2f}s')
            return (rho, solver_times)
        
        else: # not nu_max
            blocksize= len(mapping_block[nu])
            # initial values
            rho_nu = np.zeros((blocksize, chunksize), dtype = complex)
            solver_times = np.zeros(chunksize)
            
            # integrator 
            r = ode(_intfunc_block_interp).set_integrator('zvode', method = 'bdf', atol=atol, rtol=rtol)
            r.set_initial_value(rho0, t0).set_f_params(L.L0[nu], L.L1[nu], rhoff_func)
            
            if store_initial:
                rho_nu[:,0] = rho0
                solver_times[0] = t0
                n_t = 1
            else:
                n_t=0
            while r.successful() and n_t<chunksize:
                rho_step = r.integrate(r.t+dt)
                rho_nu[:, n_t] = rho_step         
                solver_times[n_t] = r.t
                n_t += 1
            
            # elapsed = time() - tstart
            # print(f'Block nu={nu}: elapsed {elapsed:.2f}s')


            return (rho_nu, solver_times)
  
        
  
    def time_evolve_chunk_parallel(self, expect_oper, chunksize = 50, progress=False, \
                        save_states=False, num_cpus=None, method='bdf', expect_per_nu=False, start_block=None\
                            ,verbose = True):
        """ 
            Parallel time evolution function
        
            Parameters
            ----------
            expect_oper : list
                List of operators for which expectation values are calculated at each time step
            chunksize : int
                Number of time steps to calculate serially in one chunk
            save_states : bool
                True: save states at all times, False: only save initial and final state
            progress : bool
                True: Enable progress bar
            num_cpus : int
                Number of CPUs for parallel computation
            method : string
                Method for integrator
                'bdf' : stiff
                'adams' : non-stiff
            expect_per_nu : bool
                True: Calculate the expectation values for each block separately
            start_block : int
                specify excitation number corresponding to block at which the time evolution should start.
                If None, start_block is set to nu_max, which gives exact result.
            verbose : bool
                True: print status messages
            
            Returns
            -------
            None
        
        Parallelize and minimize the amount of stored states. In this function, the synchronization
        between processes is done 'by hand'. Meaning we loop through chunks, and for each iteration 
        of the loop, a new parallel pool is being set up that makes use of the results from the previous chunk. 
        To minimize memory usage, only one past chunk is being saved for each nu.
        
        """
        
        if self.indices.only_numax:
            print('Parallel time evolution not implemented for only_numax. Starting serial time evolution instead...')
            self.time_evolve_block_interp(expect_oper, progress = progress, save_states=save_states, method=method)
        
        # if expect_oper == None:
        #     self.time_evolve_block(save_states=True, progress=progress)
        #     return
       

               
        num_blocks = len(self.indices.mapping_block)
        nu_max = num_blocks-1
        t0 = 0
        ntimes = round(self.tend/self.dt)+1
        
        # determine start block
        if start_block == None or start_block == nu_max:
            start_block = nu_max
        if start_block > nu_max or start_block < 0:
            raise ValueError(f"Start block exciation must be within 0 and {nu_max}.")
        if not isinstance(start_block, (int, np.integer)):
            raise ValueError('Start block excitation must be an integer.')
        if verbose:
            print(f'Start block set to {start_block}/{nu_max}.')
            print(f'Starting time evolution in chunks (parallel 2), chunk size {chunksize}...')
        tstart = time()
        
        num_evolutions = int(np.ceil(ntimes/chunksize))+nu_max # number of loops for progress bar
        # some weird edge case, if chunksize divides ntimes
        if ntimes % chunksize == 0:  
            num_evolutions += 1
        if progress:
            bar = Progress(num_evolutions, description='Time evolution total progress...', start_step=0)
        
        if save_states:
            #self.result.rho = [[] for _ in range(num_blocks)]
            self.result.rho = [ np.zeros((len(self.indices.mapping_block[i]), ntimes), dtype=complex) for i in range(num_blocks)]
            rhos_lengths = [0] * (num_blocks) # keep track of lengths of rhos, if the states are being saved

        # setup rhos as 2 times the chunksize, such that there is space for one chunk for feedforward,
        # and one chunk for calculating the future time evolution.
        saved_chunks=2 # number of stored chunks. Minimum amound is 2: One for feedforward, and one for the next time evolution
        rhos_chunk = [np.zeros((len(self.indices.mapping_block[i]),saved_chunks*chunksize), dtype=complex) for i in range(num_blocks)] 
        
        # if we want to save the exact solver time steps:
        times_chunk = np.zeros((num_blocks, saved_chunks*chunksize))
        
        # can maybe get rid of this, not sure
        #times =[np.zeros(ntimes) for _ in range(num_blocks)] # keep track of times from all blocks for interpolation (because the solver has variable time steps. Though the result is very very small, it is noticable)
        
        # keep track, where to write and where to read for each block
        write = [0] * len(self.indices.mapping_block) # zero means, write in first chunk. One means, write in second chunk
            
        # initial states for each block. Needs to be updated after each chunk calculation
        initial = [np.copy(self.rho.initial[nu]) for nu in range(num_blocks)]
        
        # if we want to save the exact solver time steps:
        initial_times = [t0] * num_blocks
        
        # initialize expectation values and time
        self.result.expect = np.zeros((len(expect_oper), ntimes), dtype=complex)
        if expect_per_nu:
            self.result.expect_per_nu = np.zeros((num_blocks, len(expect_oper),ntimes), dtype=complex)  # idx: (nu, operator, time)
        self.result.t = np.zeros(ntimes)
        self.result.t = np.arange(t0, self.tend+self.dt,self.dt)
        
        
        # make static variables global for parallel processes to use
        global L, mapping_block, t
        L = self.L
        mapping_block = self.indices.mapping_block
        t = self.result.t
        
        
        # keep track of which blocks are finished:
        finished = [0]*num_blocks

        # Loop until a chunk is outside of ntimes
        count = 0
        
        # The calculation for block nu can only be started after (nu_max-nu) chunks.
        # total computation steps: ntimes + (nu_max-1) * chunksize
        while count * chunksize - (nu_max)*chunksize <= ntimes:    # - (nu_max)*chunktimes, because the nu blocks have to be solved with a time delay      
            # setup arglist for multiprocessing pool
            arglist = [] #rho0, rhoff_func, t0,nu, chunksize, store_initial
            
            # first calculate block nu_max
            if count == 0:
                t0_numax = t0+count*chunksize*self.dt
            else: 
                t0_numax = t0+count*chunksize*self.dt - self.dt # because only in the first chunk, the initial condition does not need to be stored again. 
                                                                # The initial condition of second chunk is the last time of the previous chunk
            
            # IF we store exact solver time steps:
            t0_numax = initial_times[nu_max]                                                    
            
            # only calculate time evolution, if tend has not been reached yet
            if t0_numax < self.tend:
               # print(t0_numax)
                # In the initial chunk, save the initial state. Otherwise, don't save initial state,
                # because it has already been calculated as the final state of the previous chunk
                if count == 0:
                    save_initial = True
                else:
                    save_initial = False
                
                arglist.append((initial[nu_max], None, t0_numax, nu_max, chunksize, save_initial, self.rtol, self.atol, method))
            else:
                finished[nu_max] = 1
                

            # get the number of blocks, one can simultaneously propagate in current chunk
            # cannot exceet num_blocks, otherwise grows linearly with number of chunks
            parallel_blocks = min(count+1, num_blocks) 
            
            # repeat what has been done already for nu_max
            for nu in range(nu_max -1,nu_max - parallel_blocks, -1):
                if count - (nu_max - nu) == 0:
                    # for the first chunk, the initial value is already given, so we do not need to calculate them again
                    # that is why t0 and feedforward are shifted by one
                    t0_nu = t0 + (count - (nu_max - nu)) * chunksize*self.dt
                    feedforward = rhos_chunk[nu+1][:,write[nu] * chunksize:(write[nu]+1)*chunksize]
                    t_ff = times_chunk[nu+1,write[nu] * chunksize:(write[nu]+1)*chunksize]
                else:
                    # for all but the first chunk, the initial conditions are stored in the previous chunk,
                    # so the initial conditions are shifted by one timestep
                    t0_nu = t0 + (count - (nu_max - nu)) * chunksize*self.dt-self.dt
                    
                    end_idx = (write[nu]+1)*chunksize-1
                    start_idx = write[nu] * chunksize-1
                    if start_idx < 0 : # initial state is always the last state of previous chunk. If that chunk is written in the second half of rhos_chunk, we need to take care that a negative first index means: take the last element of the array
                        feedforward = np.concatenate((rhos_chunk[nu+1][:,-1:], rhos_chunk[nu+1][:,:end_idx]), axis=1)
                        t_ff = np.concatenate((times_chunk[nu+1, -1:], times_chunk[nu+1,:end_idx]))
                    else:
                        feedforward = rhos_chunk[nu+1][:, start_idx:end_idx]
                        t_ff = times_chunk[nu+1, start_idx:end_idx]
                        
                # IF we store exact solver time steps:
                t0_nu = initial_times[nu]

                #print(nu, count-nu_max+nu)
                if t0_nu < self.tend and not np.isclose(t0_nu, self.tend):
                    # only store initial state, if it is the very first state. Otherwise, initial state is already stored as last state in previous chunk
                    if count - (nu_max-nu) == 0:
                        save_initial = True
                    else:
                        save_initial = False
                    
                    #t_ff2 = np.linspace(t0_nu, t0_nu+chunksize*self.dt-self.dt, chunksize)
                    #print('t_ff', t_ff[0], t_ff[-1], 'ff', feedforward[:,0], feedforward[:,-1])

                    
                    feedforward_func = interp1d(t_ff, feedforward, bounds_error=False, fill_value='extrapolate', kind='cubic')
                    
                    arglist.append((initial[nu], feedforward_func, t0_nu, nu, chunksize, save_initial, self.rtol,self.atol,method))
                else:
                    finished[nu] = 1
                  
                    
            # setup parallel pool
            with Pool(processes = num_cpus) as pool:
                results = pool.map(self.evolve_nu_parallel2, arglist)
            
            
            # Think about if it is necessary, to get the exact solver time steps for the interpolation, or is it fine to just use
            # a linearly spaced t-array for all blocks
            rhos=[]
            solver_times = []
            for res in results:
                rhos.append(res[0])
                solver_times.append(res[1])
            
                
            # how many blocks are already finished?
            f = sum(finished)
            
            # loop through results
            k = 0 # index of result
            for nu in range(nu_max-f, nu_max-len(rhos)-f, -1): # nu_max is first element in rhos, therefore need to loop through in reverse order
                rhos_chunk[nu][:, write[nu] * chunksize:(write[nu]+1)*chunksize] = rhos[k]
                
                # IF we store exact solver time steps:
                times_chunk[nu, write[nu]* chunksize:(write[nu]+1)*chunksize] = solver_times[k]
                    
                k += 1
                
                initial[nu] = rhos_chunk[nu][:,(write[nu]+1) * chunksize-1]
                
                # IF we store exact solver time steps:
                initial_times[nu] = times_chunk[nu,(write[nu]+1) * chunksize-1]
                
                # get part of expectation value corresponding to nu of that chunk
                i=0
                for t_idx in range((count-(nu_max-nu))*chunksize, (count-(nu_max-nu)+1)*chunksize):
                    if t_idx >= ntimes:
                        continue
                    read_index = write[nu] * chunksize
                    
                    expect_nu = np.array(self.expect_comp_block([rhos_chunk[nu][:,read_index+ i]],nu, expect_oper)).flatten()
                    self.result.expect[:,t_idx] +=  expect_nu 
                    
                    if expect_per_nu:
                        self.result.expect_per_nu[nu, :, t_idx] = expect_nu
                    
                    # self.result.expect[:,t_idx] = self.result.expect[:,t_idx] +  (
                    #     np.array(self.expect_comp_block([rhos_chunk[nu][:,read_index+ i]],nu, expect_oper)).flatten())
                    i = i+1
                # update write variable
                write[nu] = (write[nu] + 1) % saved_chunks # if 1 make it 0, if 0 make it 1
            
            count += 1
            if progress:
                bar.update()

        elapsed = time()-tstart
        print(f'Complete {elapsed:.0f}s', flush=True)
        

        
    def time_evolve_block_interp(self,expect_oper=None, save_states=None, progress=False, \
                        method='bdf', expect_per_nu=False, start_block=None, verbose=True):
        """ Serial time evolution function
        
            Parameters
            ----------
            expect_oper : list
                List of operators for which expectation values are calculated at each time step
            save_states : bool
                True: save states at all times, False: only save initial and final state
            progress : bool
                True: Enable progress bar
            method : string
                Method for integrator
                'bdf' : stiff
                'adams' : non-stiff
            expect_per_nu : bool
                True: Calculate the expectation values for each block separately
            start_block : int
                specify excitation number corresponding to block at which the time evolution should start.
                If None, start_block is set to nu_max, which gives exact result.
            verbose : bool
                True: print status messages
            
            Returns
            -------
            None
            
        
        """      
        

               
        # store number of elements in each block
        num_blocks = len(self.indices.mapping_block)
        blocksizes = [len(self.indices.mapping_block[nu]) for nu in range(num_blocks)]
        nu_max = num_blocks -1
        t0 = 0
        ntimes = round(self.tend/self.dt)+1
        
        # determine start block
        if start_block == None or start_block == nu_max:
            start_block = nu_max
        else:
            if self.indices.only_numax:
                print('Warning: Cannot simultaneously set "only_numax" evolution and starting time evolution at block not equal to nu_max.')
                print('Setting "start_block" to nu_max = {nu_max}')
                start_block = nu_max    
                
            
            if start_block > nu_max or start_block < 0:
                raise ValueError(f"Start block exciation must be within 0 and {nu_max}.")
            if not isinstance(start_block, (int, np.integer)):
                raise ValueError('Start block excitation must be an integer.')
            if verbose:
                print(f'Start block set to {start_block}/{nu_max}.')
            
        if verbose:
            print('Starting time evolution serial block (interpolation)...')
        tstart = time()
        
        if progress:
            if self.indices.only_numax:
                bar = Progress(2*ntimes, description='Time evolution under L (only numax)...', start_step=1)
            elif start_block < nu_max:
                bar = Progress(2*(ntimes-1)*(start_block+1), description=f'Time evolution under L (start block: {start_block})...', start_step=1)
            else:
                bar = Progress(2*(ntimes-1)*num_blocks, description='Time evolution under L...', start_step=1)
        if save_states is None:
            save_states = True if expect_oper is None else False
        if not save_states and expect_oper is None:
            print('Warning: Not recording states or any observables. Only initial and final'\
                    ' states will be returned.')
                

        # set up results:
        self.result.t = np.zeros(ntimes)
        if save_states:
            self.result.rho = [np.zeros((blocksizes[nu], ntimes), dtype=complex) for nu in range(num_blocks)]
        else: # only record initial and final states
            self.result.rho = [np.zeros((blocksizes[nu], 2), dtype=complex) for nu in range(num_blocks)]
            
        if expect_oper is not None:
            self.result.expect = np.zeros((len(expect_oper), ntimes), dtype=complex)
            
            if expect_per_nu:
                self.result.expect_per_nu = np.zeros((num_blocks, len(expect_oper),ntimes), dtype=complex)  # idx: (nu, operator, time)
        
        #Record initial values
        self.result.t[0] = t0            
        
        # first calculate block nu_max. Setup integrator
        # If start_block is set by user, start with start_block
        r = ode(_intfunc).set_integrator('zvode', method = method, atol=self.atol, rtol=self.rtol, nsteps=self.nsteps)
        
        if self.indices.only_numax:
            r.set_initial_value(self.rho.initial[nu_max],t0).set_f_params(self.L.L0[0])
        else:
            # r.set_initial_value(self.rho.initial[nu_max],t0).set_f_params(self.L.L0[nu_max])
            r.set_initial_value(self.rho.initial[start_block],t0).set_f_params(self.L.L0[start_block])

        
        # temporary variable to store states
        #rhos = [ np.zeros((len(self.indices.mapping_block[i]), ntimes), dtype=complex) for i in range(num_blocks)]
        #rhos[nu][:,0] = self.rho.initial[nu]

        rho_nu = np.zeros((blocksizes[start_block], ntimes), dtype = complex)
        rho_nu[:,0] = self.rho.initial[start_block] 
        
        # using the exact solver times for the feedforward interpolation instead of using linearly spaced time array makes a (small) difference
        solver_times = np.zeros(ntimes)
        solver_times[0] = t0
    
        n_t=1
        while r.successful() and n_t < ntimes:
            rho = r.integrate(r.t+self.dt)
            self.result.t[n_t] = r.t
            solver_times[n_t] = r.t
            #rhos[nu][:,n_t] = rho
            rho_nu[:,n_t] = rho
            n_t += 1
            
            if progress:
                bar.update()
                
        # calculate nu_max part of expectation values
        if expect_oper is not None:
            # if progress:
            #     bar = Progress(ntimes, description='Calculating expectation values...', start_step=1)
                
            for t_idx in range(ntimes):
                #self.result.expect[:,t_idx] +=  np.array(self.expect_comp_block([rhos[nu][:,t_idx]],nu, expect_oper)).flatten()
                expect_nu = np.array(self.expect_comp_block([rho_nu[:,t_idx]],start_block, expect_oper)).flatten()
                self.result.expect[:,t_idx] +=  expect_nu #np.array(self.expect_comp_block([rho_nu[:,t_idx]],nu_max, expect_oper)).flatten()
                
                if expect_per_nu:
                    self.result.expect_per_nu[start_block, :, t_idx] = expect_nu

                
                if progress:
                    bar.update()
        if save_states:
            self.result.rho[start_block] = rho_nu
        else: # store initial and final states
            self.result.rho[start_block][:,0] = rho_nu[:,0]
            self.result.rho[start_block][:,1] = rho_nu[:,-1] 
            
        #self.result.t = np.arange(t0, self.tend+self.dt,self.dt)
        if self.indices.only_numax:
            elapsed = time()-tstart
            
            if verbose:
                print(f'Complete {elapsed:.0f}s', flush=True)
            return
        
        # Now, do the feed forward for all other blocks. Need different integration function, _intfunc_block_interp
        for nu in range(start_block - 1, -1,-1):           
            #rho_interp = interp1d(self.result.t, rhos[nu+1], bounds_error=False, fill_value="extrapolate") # extrapolate results from previous block
            rho_interp = interp1d(solver_times, rho_nu, bounds_error=False, fill_value="extrapolate", kind='cubic') # interpolate results from previous block, rho_nu                  
                       
            r = ode(_intfunc_block_interp).set_integrator('zvode', method = method, atol=self.atol, rtol=self.rtol, nsteps=self.nsteps)
            r.set_initial_value(self.rho.initial[nu],t0).set_f_params(self.L.L0[nu], self.L.L1[nu], rho_interp)
            
            #Record initial value
            #rhos[nu][:,0] = (self.rho.initial[nu])
            # Update rho_nu variable for current block
            rho_nu = np.zeros((blocksizes[nu], ntimes), dtype=complex)
            rho_nu[:,0] = self.rho.initial[nu]
            solver_times[0] = t0
            
            # integrate
            n_t=1
            while r.successful() and n_t<ntimes:
                rho = r.integrate(r.t+self.dt)
                #rhos[nu][:,n_t] = rho
                solver_times[n_t] = r.t
                rho_nu[:,n_t] = rho
                n_t += 1
    
                if progress:
                    bar.update()
                    
            # calculate contribution of block nu to expectation values
            if expect_oper is not None:
                for t_idx in range(ntimes):
                    #self.result.expect[:,t_idx] +=  np.array(self.expect_comp_block([rhos[nu][:,t_idx]],nu, expect_oper)).flatten()
                    expect_nu = np.array(self.expect_comp_block([rho_nu[:,t_idx]],nu, expect_oper)).flatten()
                    self.result.expect[:,t_idx] +=  expect_nu
                    
                    if expect_per_nu:
                        self.result.expect_per_nu[nu, :, t_idx] = expect_nu
                    
                    if progress:
                        bar.update()
            if save_states:
                self.result.rho[nu] = rho_nu
            else:
                self.result.rho[nu][:,0] = rho_nu[:,0]
                self.result.rho[nu][:,1] = rho_nu[:,-1] 

        elapsed = time()-tstart
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        
    
    
    
    def expect_comp_block(self, rho_list, nu, ops):
        """ Calculate expectation value of ops in block nu"""
        
        num_blocks = len(self.rho.convert_rho_block_dic)

        # converted densities matrices (different operators may have different number of
        # spins; we need a list of reduced density matrices for each number of spins)
        rhos_converted_dic = {}
        
        output = []
        for op in ops:
            # number of spins in the target rdm
            nrs = int(np.log(op.shape[0]//self.indices.ldim_p)/np.log(self.indices.ldim_s))

            if nrs not in self.rho.convert_rho_block_dic:
                raise TypeError('need to run setup_convert_rho_nrs({})'.format(nrs))

            # Only convert compressed matrices in rho
            if nrs not in rhos_converted_dic:
                rhos_converted_dic[nrs] = []
                for count in range(len(rho_list)):
                    # rho_nrs = self.rho.convert_rho_block_dic[nrs][nu].dot(rho_list[count]) # .dot
                    rho_nrs = self.rho.convert_rho_block_dic[nrs][nu]@rho_list[count]   # @
                    rho_nrs = vector_to_operator(rho_nrs)
                    rhos_converted_dic[nrs].append(rho_nrs)
            
            one_output = [] 
            rhos_converted = rhos_converted_dic[nrs]
            for count in range(len(rho_list)):
                one_output.append(expect(rhos_converted[count], op) / self.rho.scale_rho)    # added scale functionality
            output.append(one_output)
        return output
    
    
def _intfunc(t, y, L):
    """ Integration funciton for block nu_max"""
    return (L@y)
   

def _intfunc_block_interp(t,y,L0,L1,y1_func):
    """ Integration function for blocks nu < nu_max with coupling L1 to higher blocks"""
    # return (L0.dot(y) + L1.dot(y1_func(t)))
    return (L0 @ y + L1 @ y1_func(t))






#!/usr/bin/env python
import numpy as np
from itertools import product
from multiprocessing import Pool

from .util import degeneracy_spin_gamma, degeneracy_gamma_changing_block_efficient
from .util import states_compatible, permute_compatible, degeneracy_outer_invariant_optimized
from .util import _multinominal
from .util import Progress

import os
import pickle
from time import time
import scipy.sparse as sp
from itertools import permutations



class Indices:
    """Indices for and mappings between compressed and supercompressed 
    density matrices containing unique representative elements for each
    permutation invariant subspace of nspins identical spins. The 
    representative elements are chosen such that, if i=0,1,...,nspins-1
    enumerates the spins one has zeta_0 <= zeta_1 <=... is monotonically
    increasing where zeta is the simple finite pairing
    function
        zeta_i = (dim_s) * s^L_i + s^R_i
    for site (spin) i having element |s^L_i><s^R_i| with left and right 
    spin values s^L_i and s^R_i. Nominally, dim_s = 2 i.e. s^L_i, s^R_i
    = 0 or 1 (spin-1/2).

    In the compressed form, the spin indices are multiplied by the left/right
    photon part of the density matrices.
    In the supercompressed form, the (smaller) set of elements are grouped
    according to the total excitation number nu
    [can we get rid of compressed form entirely?]
    """
    def __init__(self, nspins, nphot=None,spin_dim=2, verbose=True, debug=False,\
                 save=False, index_path=None, only_numax = False):
        """

        Parameters
        ----------
        nspins : int
            Number of two-level systems
        nphot : int, optional
            Photon space dimension. If None, nphot=nspins+1
        spin_dim : int, optional
            Dimension of spin space. The default is 2 (only implemented value)
        verbose : bool, optional
            Print status messages. The default is True.
        debug : bool, optional
            If True, do not load existing file, always calculate new set of indices. The default is False.
        save : bool, optional
            If True, save spin indices for later use. The default is False.
        index_path : string, optional
            Save path for index file. The default is None.
        only_numax : bool, optional
            If true, calculate indices only for block with maximum excitation numbers. 
            This is interesting when solving a system with no losses, then we stay in the block 
            determined by initial conditions (which for sf initial conditions is nu_max). The default is False.


        Returns
        -------
        None.

        """
        
        # make some checks for validity of nspins, nphot, spin_dim
        if (not isinstance(nspins, (int, np.integer))) or nspins <= 0:
            raise ValueError("Number of spins must be integer N > 0")
        if nphot is not None and ((not isinstance(nphot, (int, np.integer))) or nphot <= 0):
            raise ValueError("Number of photon states must be integer > 0")
        if spin_dim is not None and ((not isinstance(spin_dim, (int, np.integer))) or spin_dim <= 1):
             raise ValueError("Spin dimension must be integer > 1")

            
        if nphot is None:
            nphot = nspins + 1
            if verbose:
                print('Photon dimension set to ', nphot)
        if spin_dim is None:
            spin_dim = 2 # spin 1/2 system
            if verbose:
                print('Spin dimension set to ', spin_dim)
            
        self.nspins, self.ldim_p, self.ldim_s = nspins, nphot, spin_dim
        self.indices_elements = []
        self.indices_elements_inv = {}
        self.mapping_block = []
        self.elements_block = []
        self.difference_block_inv = []
        self.coupled_photon_block = []
        
        self.only_numax = only_numax
        
        self.verbose = verbose
        
        # loading/saving paths
        if index_path is None:
            index_path = 'data/indices/'
        if not os.path.isdir(index_path):
            os.makedirs(index_path)
            print('Created directory ', index_path)
        filename = f'indices_Ntls{self.nspins}_Nphot{self.ldim_p}_spindim{self.ldim_s}.pkl'
        fname_numax = f'indices_numax_Ntls{self.nspins}_Nphot{self.ldim_p}_spindim{self.ldim_s}.pkl' # for the case only_numax =True
        
        
        if debug is False:
            if only_numax: # check data/indices/numax
                if os.path.isdir(index_path+'numax/'):
                    index_files = os.listdir(index_path + 'numax/')
                    if any([f == fname_numax for f in index_files]):
                        self.load(index_path + 'numax/' + fname_numax)
                        return
                else:
                    os.makedirs(index_path + 'numax/')
                
            # check if an object with the same arguments already exists in data/indices/ folder
            index_files = os.listdir(index_path)
            if (any([f == filename for f in index_files])):
                self.load(index_path+filename)
                return
            
        # debug true -> always calculate spin indices anew, or if not save file is found
        # setup indices
        if verbose:
            print(f'Setting up spin indices with nspins={self.nspins},spin_dim={self.ldim_s}...', flush=True)
        t0 = time()
        self.list_equivalent_elements()
        elapsed = time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)

        # setup mapping block
        if verbose:
            if only_numax:
                print(f'Running setup indices block (only nu_max) with nspins={self.nspins},nphot={self.ldim_p}...', flush=True)
            else:
                print(f'Running setup indices block with nspins={self.nspins},nphot={self.ldim_p}...', flush=True)
        t0 = time()
        self.setup_mapping_block()
        elapsed = time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        
        if save:
            # export for future use, if save is true
            if only_numax:
                self.export(index_path + 'numax/' + fname_numax)
            else:
                self.export(index_path + filename)
                
        for nu in range(len(self.mapping_block)):
            assert len(self.mapping_block[nu]) == len(self.elements_block[nu])
                
    
    def list_equivalent_elements(self):
        """Generate a list of elements [left, right] representing all permutation
        distinct spin elements |left><right|"""
        #get minimal list of left and right spin indices - in combined form (i.e. as a list of zetas)
        #Generate list of all unique zeta strings in reverse lexicographic order
        all_zetas = [np.zeros(self.nspins, dtype=int)]
        # e.g. spin-1/2 -> s = 0,1 and zeta = 0, 1, 2, 3 = 2s_L + s_R
        max_zeta = self.ldim_s**2 - 1 # (d_s-1)(d_s+1)
        self.recurse_lexi(all_zetas, 0, max_zeta)
       
        for count, zetas in enumerate(all_zetas):
            left, right = self._to_hilbert(zetas)
            spin_element = np.concatenate((left,right))
            self.indices_elements.append(spin_element) # elements are stored as np.array of spin values for left (bra) then for right (ket)
            self.indices_elements_inv[tuple(zetas)] = count # mapping from combined form to index of indices_elements

    def recurse_lexi(self, all_zetas, current_index, max_zeta):
        """Generate successive strings of zetas, appending to all_zetas list"""
        previous_element = all_zetas[-1]
        for zeta in range(1, max_zeta+1):
            next_element = np.copy(previous_element)
            next_element[current_index] = zeta
            all_zetas.append(next_element)
            if current_index < self.nspins-1:
                self.recurse_lexi(all_zetas, current_index+1, zeta) 

    
    def setup_mapping_block(self):
        """
        Generate mapping between reduced representation of density matrix and
        the block structure, which is grouped in different numbers of total excitations
        of photons + spins. Note: 0 in spin array means spin up!
        For now, nu_max (maximum excitation) is set to nspins, because the initial 
        condition is always all spins up and zero photons in the cavity.
        
        Structure of mapping_block = [ [indices of nu=0] , [indices of nu=1], ... [indices of nu_max] ]
        
        IF self.only_numax == True -> only get the block with nu_max. This block has each
        element of self.indices_elements exactly once, which is because any spin configuration is allowed (in max. excitation block)
        and the photon numbers then get just matched such that left and right excitations equal nu_max.

        """     
        num_elements = len(self.indices_elements) # number of distinct spin states
        nu_max = self.nspins # maximum excitation number IF initial state is all spins up and zero photons
                             # TODO: depending on the initial state of spins and photons, adapt nu_max properly 
        
        self.mapping_block = [ [] for _ in range(nu_max+1)] # list of nu_max+1 empty lists
        self.elements_block = [ [] for _ in range(nu_max+1)]
        self.elements_block_inv = [ {} for _ in range(nu_max+1)]
        difference_block = [ [] for _ in range(nu_max+1)] # used to create difference_block_inv
        self.difference_block_inv = {nu:[] for nu in range(nu_max+1)} # used for rdm conversion calculations,
        # contains index and difference between left and right spin states for each spin element
        self.coupled_photon_block = [{} for _ in range(nu_max+1)] # at each nu, dictionary of key-values 
        # where key is a photon state as a tuple i.e. |p1><p2| -> (p1,p2) and
        # values is a pair of lists [same_block, below_block], where same_block
        # is to contain all the photon states (tuples) that may couple to |p1><p2| at nu, 
        # and below_block those states (tuples) that couple in the block below (via decay)
        # [Construction lends to needing the state in excitation below rather than above]
        coupled_photon_counts = {} # dictionary, elements under key (p1,p2) describe all photon 
        # states (tuples) which |p1><p2| couples to
        
        # for initial state calculation:
        # self.elements_photon_diag = {nu: [] for nu in range(nu_max+1)}

        def get_coupled_counts(p1, p2):
            """List all states coupled to |p1><p2| by physical processes
            Returns [ same_block_counts, below_block_counts ], where same and below counts
            indicate states in same and one lower excitation block, respectively
            """
            eye = (p1,p2) # act identically
            dleft = (max(p1-1,0),p2) # destroy left (e.g. a sigma^+ in LM coupling)
            cleft = (min(p1+1,nu_max),p2) # create left
            dright = (p1,min(p2+1,nu_max)) # destroy right
            cright = (p1,max(p2-1,0)) # create right
            same_block_counts = list(set((eye, dleft, cleft, dright, cright))) # set removes any duplicates
            below_block_counts = [eye] # photon count stays the same, spin decays
            if p1 > 0 and p2 > 0: # e.g. |p1><p2|=|0><1| does NOT decay to |-1><0| !
                below_block_counts.append((p1-1, p2-1)) # photon decay
            return [ same_block_counts, below_block_counts ]
        
        # populate coupled_photon_counts dictionary and initialise coupled_photon_block 
        for count_tuple in product(range(nu_max+1), range(nu_max+1)):
            coupled_photon_counts[count_tuple] = get_coupled_counts(*count_tuple)
            for nu in range(max(count_tuple), nu_max+1):
                self.coupled_photon_block[nu][count_tuple] = [[],[]]

        for count in range(num_elements): # loop through spin states
            element = self.indices_elements[count]  # spin-element in compressed form
            left = element[0:self.nspins]   # left spin indices
            right = element[self.nspins:2*self.nspins] # right spin indices
            zetas = 2 * left + right 
            m_left = self.nspins-sum(left) # excitations in left spins
            m_right = self.nspins-sum(right) # excitations in right spins
            num_diff = sum(left != right)
            nu_min = max(m_left, m_right) # can't have fewer than m_left+0 photons (or m_right+0photons) excitations
            
            # nu_max_loop can at most be nu_max (because no element has more excitations than nu_max)
            # but in the case where nphot != ntls+1, it could be that nu_max_loop < nu_max, because
            # the photon space is restricted. Then, nu_max_loop = maximum excitation possible with left or right spin index plus photon excitations.
            nu_max_loop = min(max(m_left, m_right) + self.ldim_p-1, nu_max)
            
            if self.only_numax == True:
                nu_min = nu_max
            
            for nu in range(nu_min, nu_max_loop+1):
                count_p1 = nu - m_left
                count_p2 = nu - m_right
                if count_p1 >= self.ldim_p or count_p2 >= self.ldim_p: # disregard states, where photon number in left or right index exceeds photon space dimension
                    continue
                
                element_index = self.ldim_p*num_elements*count_p1 + num_elements*count_p2 + count
                el = np.concatenate(([count_p1], left, [count_p2], right))
                self.mapping_block[nu].append(element_index)
                self.elements_block[nu].append(el)
                # can't use inverse here we sort the mapping block below - instead
                # note difference and fill difference_block_inv AFTER sorting
                # below. 
                #self.difference_block_inv[num_diff].append((nu, len(self.mapping_block[nu])))
                difference_block[nu].append(num_diff)
                
                # for general initial state: store indices of elements, which have the same 
                # number of photons in left and right state
                # if count_p1 == count_p2:
                #     self.elements_photon_diag[count_p1].append((nu, len(self.mapping_block[nu])-1))

        # Re-order to match that of earlier implementations
        if self.only_numax == True:
            nu_min = nu_max
        else:
            nu_min = 0
        for nu in range(nu_min,nu_max+1): # need to go through all nu here
            # zip-sort-zip - a personal favourite Python One-Liner
            self.mapping_block[nu], self.elements_block[nu], \
            difference_block[nu] =\
                    zip(*sorted(zip(self.mapping_block[nu],
                                    self.elements_block[nu],
                                    difference_block[nu])))
            # have to populate difference and coupled photon indices AFTER sort
            for i, num_diff in enumerate(difference_block[nu]):
                self.difference_block_inv[num_diff].append((nu, i))
            for i, element in enumerate(self.elements_block[nu]):
                count_tuple = (element[0], element[self.nspins+1])
                same_block_counts, below_block_counts = coupled_photon_counts[count_tuple]
                for target_tuple in same_block_counts:
                    try:
                        self.coupled_photon_block[nu][target_tuple][0].append(i)
                    except KeyError:
                        pass # target_tuple does not exist in this excitation block!
                        # e.g. |2><1| couples to |3><1| (create photon
                        # in left state) at nu=3 but NOT nu=2
                if nu == 0:
                    # at lowest block already
                    continue
                for target_tuple in below_block_counts:
                    try:
                        # state at target_tuple in block below couples to count_tuple, so add index
                        self.coupled_photon_block[nu-1][target_tuple][1].append(i)
                    except KeyError:
                        pass # e.g. |2><2| at nu=3 spin decay to nu=2 is valid
                        # but |2><2| at nu=2 spin decay to nu=1 is NOT (since
                        # |2><2| is not in the nu=1 block)
    
    def _to_hilbert(self, combined):
        """Convert zeta-string to |left> and <right| spin values"""
        right = combined % self.ldim_s
        left = (combined - right)//self.ldim_s
        return left, right


    def export(self, filepath):
        if self.verbose:
            print(f'Storing Indices for later use in {filepath} ...')
        t0 = time()
        with open(filepath, 'wb') as handle:
            pickle.dump(self, handle)
        elapsed = time() - t0
        if self.verbose:
            print(f'Storing complete {elapsed:.1f}s')
            


    def load(self, filepath):
        if self.verbose:
            print(f'Loading indices file {filepath} ...')
        t0 = time()
        with open(filepath, 'rb') as handle:
            indices_load = pickle.load(handle)
        self.__dict__ = indices_load.__dict__
        elapsed = time() - t0
        # do some checks
        # at least tell user what they loaded
        if not hasattr(self, 'only_numax'):
            self.only_numax = False # if the indices file loaded is an old one that does not have the attribute 'only_numax' yet, then set it to false.

        if self.verbose:            
            if self.only_numax:
                print(f'Loaded index file with ntls={self.nspins}, nphot={self.ldim_p}, spin_dim={self.ldim_s} (only nu_max) {elapsed:.1f}s')
            else:
                print(f'Loaded index file with ntls={self.nspins}, nphot={self.ldim_p}, spin_dim={self.ldim_s} {elapsed:.1f}s')
        
        
        
      
    # DEBUGGING FUNCTIONs   
        
    def print_elements(self, numax=None):
        """ Print elements in each block """
        if numax is None:
            numax = len(self.mapping_block)
        for nu in range(numax):
            print(30* '-', f'nu={nu}',30* '-')
            for i in range(len(self.mapping_block[nu])):
                left = self.elements_block[nu][i][1:self.nspins+1]
                right = self.elements_block[nu][i][self.nspins+2:]
                xis = 2*left + right
                print(f'{i}:',self.elements_block[nu][i], xis)
                
    def print_spins(self):
        """ Print all spin elements"""
        print('Printing spins: (0=up, 1=down)')
        count = 0
        for element in self.indices_elements:
            left = element[0:self.nspins]
            right = element[self.nspins+1:]
            xi = 2*left + right
            print(f'{count}: {element}, xi: {xi}')
            count +=1
                
    def element_count(self):
        """ Print the number of elements in each block """
        sizeL0 = np.array([len(block)**2 for block in self.mapping_block])
        sizeL1 = [len(self.mapping_block[nu]) * len(self.mapping_block[nu+1]) for nu in range(len(self.mapping_block)-1)]
        sizeL1.append(0)
        sizeL1 = np.array(sizeL1)
        
        loops_photon_trick = np.zeros(len(self.mapping_block))
        for nu in range(len(self.mapping_block)):
            count_nu = self.coupled_photon_block[nu]
            for key in count_nu:
                loops_photon_trick[nu] += len(count_nu[key][0]) + len(count_nu[key][1])
    
        
        print('Number of elements:', [len(block) for block in self.mapping_block])
        print('Size L0:', sizeL0)
        print('Size L1:', sizeL1)
    
                        


class BlockL:
    """Calculate Liouvillian basis block form. As a requirement, the Master
    Equation for the system must have weak U(1) symmetry, such that density-matrix
    elements couple to density matrix elements that differ in the total number of
    excitations (photons + spin) by at most one. We label the total number of excitation
    by nu.
    
    The Liouvillian block structure consists of 2 sets of matrices: L0 and L1.
    L0 contains square matrices, which couple density matrix elements from one
    block to the same block (preserve total excitation number).
    L1 contains matrices, which couple elements from block nu to elements in block
    nu+1. There are nu_max matrices in L0, and nu_max-1 matrices in L1, because
    the block of nu_max cannot couple to any higher nu.
    
    In this first version, we calculate the Liouvillian for the Dicke model with
    photon loss L[a], individual dephasing L[sigma_z] and individual exciton loss
    L[sigma_m], as well as a Hamiltonian of the form 
    H = wc*adag*a + sum_k { w0*sigmaz_k  + g*(a*sigmap_k + adag*sigmam_k) }
    This means we have 6 parameters, i.e. we need a basis of 6 elements. These
    we can then later scale by the parameters and add up, to get the desired system
    Liouvillian
    
    This is convenient, because we can calculate the basis for a given N, Nphot
    once, and reuse them for all different values of the dissipation rates or energies
    in the hamiltonian.
    
    
    IF indices.only_numax is true, L0_basis and L0 have only one entry, and L1_basis and L1 have no entries
    (Not like in the spin indices case, where mapping_block e.g. has num_blocks - 1 empty arrays before nu_max block)
    
    L calculation in the case indices.only_numax = True only for
        parallel=0
        parallel=1
    """
    def __init__(self, indices, parallel=0,num_cpus=None, debug=False, save=True,\
                 progress=False, liouv_path=None,verbose=True):
        """
        

        Parameters
        ----------
        indices : setup.Indices
            Indices object
        parallel : int, optional
            Liouvillian construction:
                0: serial, 1: parallel. The default is 0.
        num_cpus : int, optional
            Number of CPUs. The default is None.
        debug : bool, optional
            If True, do not load existing file, always calculate Liouvillian from scratch. The default is False.
        save : bool, optional
            If True, save Liouvillian for later use. The default is True.
        progress : bool, optional
            Progress bar. The default is False.
        liouv_path : string, optional
            Save path for Liouvillian files. The default is None.
        verbose : bool, optional
            Print status messages. The default is True.

        Returns
        -------
        None.

        """
        
        if indices.only_numax:
            if parallel > 1:
                raise ValueError(f'For calculation with only nu_max, use parallel=0 or parallel=1. Parallel = {parallel} is not supported.')
                
        
        # initialisation
        # L0 : does not change excitation number
        # L1 : changes excitation number 
        self.L0_basis = {'sigmaz': [],
                         'sigmam': [],
                         # 'sigmam_collective': [],
                         'a': [],
                         'ad_sigmam' : [],
                         'a_sigmap'  : [],
                         'H_n': [],
                         'H_sigmaz': [],
                         'H_g': []}
        self.L1_basis = {'sigmam': [],
                         # 'sigmam_collective':[],
                         'a': []}
        self.num_cpus = num_cpus
        self.verbose = verbose
        
        
        if liouv_path is None:
            liouv_path = 'data/liouvillians/'
        if not os.path.exists(liouv_path):
            os.makedirs(liouv_path)
            print('Created directory ', liouv_path)
            
        filename = f'liouvillian_TC_Ntls{indices.nspins}_Nphot{indices.ldim_p}_spindim{indices.ldim_s}.pkl'
        fname_numax = f'liouvillian_TC_numax_Ntls{indices.nspins}_Nphot{indices.ldim_p}_spindim{indices.ldim_s}.pkl'
        
        if debug is False:
            if indices.only_numax: # check, if we only need to load nu_max
                if os.path.isdir(liouv_path + 'numax/'):
                    liouv_files= os.listdir(liouv_path + 'numax/')
                    if (any([f == fname_numax for f in liouv_files])):
                        self._load(liouv_path + 'numax/'+fname_numax, indices)
                        return
                else:
                    os.makedirs(liouv_path + 'numax/')
             
            # check if an object with the same arguments already exists in data/liouvillian/ folder
            liouv_files = os.listdir(liouv_path)
            if (any([f == filename for f in liouv_files])):
                self._load(liouv_path+filename, indices)
                return
            
        # if not, calculate them
        t0 = time()
        pname = {0:'serial', 1:'parallel (lines)', 2:'parallel (blocks) (WARNING: memory inefficient, testing only)',3:'ray module (lines)', 4: 'ray module (blocks)'}
        pfunc = {0: self.setup_L_block_basis, 1: self.setup_L_block_basis_parallel,
                 2: self.setup_L_block_basis_parallel2
                 }
        try:
            if verbose:
                if indices.only_numax:
                    print(f'Calculating normalised Liouvillian (only nu_max) {pname[parallel]}...')
                else:
                    print(f'Calculating normalised Liouvillian {pname[parallel]}...')
            pfunc[parallel](indices, progress)
        except KeyError as e:
            print('Argument parallel={parallel} not recognised')
            raise e
        elapsed = time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        
        if save:
            # export normalized Liouvillians for later use, if save is true
            if indices.only_numax:
                self.export(liouv_path+'numax/'+fname_numax)
            else:
                self.export(liouv_path+filename)
        
        
    
    def export(self, filepath):
        if self.verbose:
            print(f'Storing Liouvillian basis for later use in {filepath} ...', flush=True)
        t0 = time()
        with open(filepath, 'wb') as handle:
            pickle.dump(self, handle)
        elapsed = time() - t0
        if self.verbose:
            print(f'Storing complete {elapsed:.1f}', flush=True)
            
    def _load(self, filepath,ind):
        t0 = time()
        if self.verbose:
            if ind.only_numax:
                print(f'Loading Liouvillian basis (only nu_max) file with ntls={ind.nspins}, nphot={ind.ldim_p}, spin_dim={ind.ldim_s} from file {filepath} ...', flush=True)
            else:
                print(f'Loading Liouvillian basis file with ntls={ind.nspins}, nphot={ind.ldim_p}, spin_dim={ind.ldim_s} from file {filepath} ...', flush = True)

        
        with open(filepath, 'rb') as handle:
            L_load = pickle.load(handle)
        if (not hasattr(L_load.indices, 'only_numax')):
            loaded_only_numax = False # If loaded L does not have attribute only_numax, set to False.
        else:
            loaded_only_numax = L_load.indices.only_numax
        if loaded_only_numax == False and ind.only_numax == True:
            # In this case, we loaded a full L basis, but only need the largest block!
            for names in L_load.L0_basis:
                # print(L_load.L0_basis['sigmam_collective'])
                self.L0_basis[names] = [L_load.L0_basis[names][-1]]
                self.L1_basis = []
        else:
        
            self.L0_basis = L_load.L0_basis
            self.L1_basis = L_load.L1_basis
        
        elapsed = time() - t0
        # at least tell user what they loaded
        if self.verbose:
            print(f'Loading complete {elapsed:.1f}', flush=True)

    
    @staticmethod    
    def sparse_constructor_dic(shape):
        # (data, (coords_x, coords_y)
        return {'data':[], 'coords':[[],[]], 'shape':shape}
    @staticmethod
    def new_entry(L_dic, name, count_in, count_out, data):
        # function to add data and coords to target L dictionary and name
        
        # check first, if an entry at those coordinates already exists
        coord_tuples = [(L_dic[name]['coords'][0][i],L_dic[name]['coords'][1][i]) for i in range(len(L_dic[name]['coords'][0]))]
        if (count_in, count_out) in coord_tuples:
            idx = coord_tuples.index((count_in, count_out))
            L_dic[name]['data'][idx] += data
        
        else: # coords (count_in, count_out) do not yet exist, then just append them
            L_dic[name]['data'].append(data)
            L_dic[name]['coords'][0].append(count_in)
            L_dic[name]['coords'][1].append(count_out)
        
    
    def setup_L_block_basis(self, indices, progress):
       """ Calculate Liouvillian basis in block form, serial version"""
       num_blocks = len(indices.mapping_block)
       
       if progress:
           num_elements = sum([len(indices.mapping_block[nu]) for nu in range(num_blocks)])
           bar = Progress(num_elements, 'Calculate L basis...')
       
       #------------------------------------------------------
       # First, get L0 part -> coupling to same block, 
       #------------------------------------------------------
       
       # loop through all elements in block structure

       if indices.only_numax:
           nu_min = num_blocks -1  # num_blocks = nu_max + 1
       else:
           nu_min = 0
       
       # Loop through all blocks
       for nu_element in range(nu_min, num_blocks):
           current_blocksize = len(indices.mapping_block[nu_element])
           # setup the Liouvillians for the current block
           names = ['sigmaz', 'sigmam', 'a', 'H_n', 'H_sigmaz',\
                    'H_g','ad_sigmam','a_sigmap'] # keys for the individual terms in Liouvillian
           names = list(self.L0_basis.keys())
           L0_new ={name:self.sparse_constructor_dic((current_blocksize, current_blocksize)) for name in names}
           if nu_element < num_blocks-1:
               next_blocksize = len(indices.mapping_block[nu_element+1])
               # Liouvillian terms coupling to next block
               names = ['sigmam', 'a']
               names = list(self.L1_basis.keys())
               
               L1_new ={name:self.sparse_constructor_dic((current_blocksize, next_blocksize)) for name in names}
           
           # Loop through all elements in one block

           for count_in in range(current_blocksize):
               if progress:
                   bar.update()
               # get element, of which we want the time derivative
               element = indices.elements_block[nu_element][count_in]
               left = element[0:indices.nspins+1] # left state, first index is photon number, rest is spin states
               right = element[indices.nspins+1:2*indices.nspins+2] # right state
               
               photon_tuple = (left[0], right[0])
               # Loop through all elements in the same block WITH compatible photon counts. Note: For L[ad_sigm], we need to adapt this, because new couplings are possible
               coupled_counts_nu = indices.coupled_photon_block[nu_element][photon_tuple][0]
               # print('WARNING: Serial L construction does not use photon count trick to calculate L[ad*sigmam]')
               for count_out in coupled_counts_nu:
                   
                   # get "to couple" element, that contributes to time derivative of "element"
                   element_to_couple = indices.elements_block[nu_element][count_out]
                   left_to_couple = element_to_couple[0:indices.nspins+1]
                   right_to_couple = element_to_couple[indices.nspins+1:2*indices.nspins+2]
                   
                   #-----------------------------
                   # get Liouvillian elements
                   #-----------------------------
                   
                   right_equal = (right_to_couple == right).all()
                   left_equal = (left_to_couple == left).all()      

                   # Diagonal parts

                   if left_equal and right_equal: 
                       # L0 part from Hamiltonian
                       s_down_right = sum(right[1:])
                       s_down_left = sum(left[1:])
                       self.new_entry(L0_new, 'H_n', count_in, count_out, -1j * (left[0]-right[0]))
                       self.new_entry(L0_new, 'H_sigmaz', count_in, count_out, 1j*(s_down_left-s_down_right))
                       
                       # L0 part from L[sigmam] -> -sigmap*sigmam*rho - rho*sigmap*sigmam
                       # make use of the fact that all spin indices contribute only, if left and right spin states in sigma^+sigma^- are both up
                       # also make use of the fact that sigma^+sigma^- is diagonal, so the two terms rho*sigma^+sigma^- and sigma^+sigma^-*rho are equal
                       deg_right = degeneracy_spin_gamma(right_to_couple[1:indices.nspins+1], right[1:indices.nspins+1]) # degeneracy: because all spin up elements contribute equally
                       deg_left = degeneracy_spin_gamma(left_to_couple[1:indices.nspins+1], left[1:indices.nspins+1])
                       self.new_entry(L0_new, 'sigmam', count_in, count_out,  - 1/2 * (deg_left+deg_right))
                       
                       
                       # L0 part from L[sigmaz] -> whole dissipator
                       # Left and right states must be equal, because sigmaz is diagonal in the spins.
                       equal = (left[1:indices.nspins+1] == right[1:indices.nspins+1]).sum()
                       self.new_entry(L0_new, 'sigmaz', count_in, count_out, 2*(equal - indices.nspins))
                       
                       # L0 part from L[a]     -> -adag*a*rho - rho*adag*a
                       self.new_entry(L0_new, 'a', count_in, count_out, -1/2*(left[0] + right[0]))
                   

                   # offdiagonal parts; from commutator part of H_g and from Lindbladian terms X*rho or rho*X
                   elif(states_compatible(right, right_to_couple)): 
                        # if they are compatible, permute left_to_couple appropriately for proper H element
                        left_to_couple_permute = np.copy(left_to_couple)
                        if not right_equal:
                            # if they are compatible but not equal, we need to permute left_to_couple appropriately, to get correct matrix element of H
                            left_to_couple_permute[1:] = permute_compatible(right[1:],right_to_couple[1:],left_to_couple[1:])
                            
                        # FIRST: H_g part of commutator -i * H_g * rho
                        # Now first check, if the matrix element is nonzero. This is the case, if all the spins but one match up.
                        if (left[1:]==left_to_couple_permute[1:]).sum() == indices.nspins-1:
                        
                            deg = degeneracy_outer_invariant_optimized(left[1:], right[1:], left_to_couple_permute[1:]) # degeneracy from simulatneous spin permutations, which leave outer spins invariant
                            # check if photon number in left state increases or decreases and
                            # if all but one spin agree, and that the spin that does not agree is down in right and up in right_to_couple
                            left_photon_diff = left[0] - left_to_couple[0]
                            left_spin_sum_diff = sum(left[1:])-sum(left_to_couple[1:])
                            if left_photon_diff == 1 and left_spin_sum_diff == 1: # need matrix element of adag*sigmam
                                self.new_entry(L0_new, 'H_g', count_in, count_out, - 1j*deg * np.sqrt(left[0]))
                            elif left_photon_diff == -1 and left_spin_sum_diff == -1 : # need matrix element of a*sigmap
                                self.new_entry(L0_new, 'H_g', count_in, count_out, - 1j*deg * np.sqrt(left[0]+1))
                        
                               
                   elif(states_compatible(left, left_to_couple)):            
                        # if they are compatible, permute right_to_couple appropriately for proper H element
                        right_to_couple_permute = np.copy(right_to_couple)
                        if not left_equal:
                            right_to_couple_permute[1:] = permute_compatible(left[1:],left_to_couple[1:],right_to_couple[1:])
                            
                        # FIRST: H_g part of commutator i * rho * H_g
                        # Now first check, if the matrix element is nonzero. This is the case, if all the spins but one match up.
                        if (right[1:]==right_to_couple_permute[1:]).sum() == indices.nspins-1:
                            deg = degeneracy_outer_invariant_optimized(left[1:], right[1:], right_to_couple_permute[1:])
                            # check if photon number in right state increases or decreases and
                            # if all but one spin agree, and that the spin that does not agree is down in right and up in right_to_couple
                            right_photon_diff = right[0] - right_to_couple[0]
                            right_spin_sum_diff = sum(right[1:])-sum(right_to_couple[1:])
                            if right_photon_diff == 1 and right_spin_sum_diff == 1: # need matrix element of a*sigmap
                                self.new_entry(L0_new, 'H_g', count_in, count_out,  1j*deg * np.sqrt(right[0]))
                            elif right_photon_diff == -1 and right_spin_sum_diff == -1: # need matrix element of adag*sigmam
                                self.new_entry(L0_new, 'H_g', count_in, count_out,  1j*deg * np.sqrt(right[0]+1))
                            

                   
               if nu_element == num_blocks -1: # no L1 part for highest block
                   continue
                
               # Now get L1 part -> coupling from nu_element to nu_element+1 loop through matrix
               # elements in the next block WITH compatible photon counts (only photon number 
               # unchanged -> spin decay or photon number decreased by one -> photon decay)
               coupled_counts_nu_plus = indices.coupled_photon_block[nu_element][photon_tuple][1]
               for count_out in coupled_counts_nu_plus:    
                   
                   # get "to couple" element
                   element_to_couple = indices.elements_block[nu_element+1][count_out]
                   left_to_couple = element_to_couple[0:indices.nspins+1]
                   right_to_couple = element_to_couple[indices.nspins+1:2*indices.nspins+2]
                   left_to_couple_spins = left_to_couple[1:]
                   right_to_couple_spins = right_to_couple[1:]
                   
                   #---------------------------------
                   # get Liouvillian elements
                   #--------------------------------
                   
                   # L1 part from L[sigmam] -> sigmam * rho * sigmap
                   # Photons must remain the same

                   if (left[0] == left_to_couple[0] and right[0] == right_to_couple[0]):
                       # we have to compute matrix elements of sigma^- and sigma^+. Therefore, check first if 
                       # number of spin up in "right" and "right_to_couple" as well as "left" and "left_to_coupole" vary by one
                       if (sum(left[1:]) - sum(left_to_couple[1:]) == 1) and (sum(right[1:]) - sum(right_to_couple[1:]) == 1):       
                           # Get the number of permutations, that contribute.                             
                           deg = degeneracy_gamma_changing_block_efficient(left[1:], right[1:], left_to_couple[1:], right_to_couple[1:])                
                           self.new_entry(L1_new, 'sigmam', count_in, count_out, deg)

                   
                   # L1 part from L[a] -> a * rho* adag
                   # since spins remain the same, first check if spin states match
                   # if spins match, then the element can couple, because we are looping through the block nu+1. Therefore
                   # the coupled-to-elements necessarily have one more excitation, which for this case is in the photon state.
                   if (left[1:] == left_to_couple[1:]).all() and (right[1:]==right_to_couple[1:]).all():
                       self.new_entry(L1_new, 'a', count_in, count_out,  np.sqrt((left[0]+1)*(right[0] + 1)))
             
            
           # append new blocks to the basis as sparse matrices (CSR format)
           for name in self.L0_basis:
               Lnew = L0_new[name]
               data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
               self.L0_basis[name].append(sp.coo_matrix((data, coords), shape=shape).tocsr())
           
           if nu_element < num_blocks-1: 
               for name in self.L1_basis:
                   Lnew = L1_new[name]
                   data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
                   self.L1_basis[name].append(sp.coo_matrix((data,coords), shape=shape).tocsr())
       
    
                
    # functions for parallelization
    @staticmethod
    def calculate_L0_line(args_tuple):
        """ Calculate L0 part of element count_in in block nu_element """
        global elements_block, new_entry, sparse_constructor_dic, coupled_photon_block_nu
        
        nu_element, count_in = args_tuple

        current_element_block = elements_block[nu_element]
        current_blocksize = len(current_element_block)
        # get element, of which we want the time derivative
        element = current_element_block[count_in]
        left = element[0:nspins+1] # left state, first index is photon number, rest is spin states
        right = element[nspins+1:2*nspins+2] # right state
    
        
        # initialize Liouvillian rows for element count_in
        names = ['sigmaz', 'sigmam', 'a', 'H_n', 'H_sigmaz', 'H_g']
        L0_line = {name:sparse_constructor_dic((current_blocksize, current_blocksize)) for name in names}
        new_entry_func = lambda name, count_out, val: new_entry(L0_line, name, count_in, count_out, val)
        
        photon_tuple = (left[0], right[0])
        # Loop through all elements in the same block WITH compatible photon counts
        coupled_counts_nu = coupled_photon_block_nu[photon_tuple][0]
        for count_out in coupled_counts_nu:
            # get "to couple" element
            element_to_couple = current_element_block[count_out]
            left_to_couple = element_to_couple[0:nspins+1]
            right_to_couple = element_to_couple[nspins+1:2*nspins+2]
            
            #-----------------------------
            # get Liouvillian elements
            #-----------------------------
           
            # Diagonal part
            if (right_to_couple == right).all() and (left_to_couple == left).all():
                # L0 part from Hamiltonian
                s_down_right = sum(right[1:])
                s_down_left = sum(left[1:])
                new_entry_func('H_n', count_out, -1j * (left[0]-right[0]))
                new_entry_func('H_sigmaz', count_out, 1j*(s_down_left-s_down_right))
                
                # L0 part from L[sigmam] -> -sigmap*sigmam*rho - rho*sigmap*sigmam
                # make use of the fact that all spin indices contribute only, if left and right spin states in sigma^+sigma^- are both up
                # also make use of the fact that sigma^+sigma^- is diagonal, so the two terms rho*sigma^+sigma^- and sigma^+sigma^-*rho are equal
                deg_right = degeneracy_spin_gamma(right_to_couple[1:nspins+1], right[1:nspins+1]) # degeneracy: because all spin up elements contribute equally
                deg_left = degeneracy_spin_gamma(left_to_couple[1:nspins+1], left[1:nspins+1])
                new_entry_func('sigmam', count_out, - 1/2 * (deg_left+deg_right))
                
                # L0 part from L[sigmaz] -> whole dissipator
                # Left and right states must be equal, because sigmaz is diagonal in the spins.
                equal = (left[1:nspins+1] == right[1:nspins+1]).sum()
                new_entry_func('sigmaz', count_out,  2*(equal - nspins))
                    
                # L0 part from L[a]     -> -adag*a*rho - rho*adag*a
                new_entry_func('a', count_out, -1/2*(left[0] + right[0]))
            
            # offdiagonal parts
            elif(states_compatible(right, right_to_couple)):
                 # if they are compatible, permute left_to_couple appropriately for proper H element
                 left_to_couple_permute = np.copy(left_to_couple)
                 if not (right_to_couple == right).all():
                     # if they are compatible but not equal, we need to permute left_to_couple appropriately, to get correct matrix element of H
                     left_to_couple_permute[1:] = permute_compatible(right[1:],right_to_couple[1:],left_to_couple[1:])
                     
                 # Now first check, if the matrix element is nonzero. This is the case, if all the spins but one match up.
                 if (left[1:]==left_to_couple_permute[1:]).sum() != nspins-1:
                     continue
                 
                 deg = degeneracy_outer_invariant_optimized(left[1:], right[1:], left_to_couple_permute[1:]) # degeneracy from simulatneous spin permutations, which leave outer spins invariant
                 # check if photon number in left state increases or decreases and
                 # if all but one spin agree, and that the spin that does not agree is down in right and up in right_to_couple
                 if (left[0] - left_to_couple[0]) == 1 and sum(left[1:])-sum(left_to_couple[1:]) == 1: # need matrix element of adag*sigmam
                     new_entry_func('H_g', count_out, - 1j*deg * np.sqrt(left[0]))

                 elif (left[0] - left_to_couple[0] == -1) and sum(left[1:])-sum(left_to_couple[1:]) == -1 : # need matrix element of a*sigmap
                     new_entry_func('H_g', count_out,- 1j*deg * np.sqrt(left[0]+1))
                        
            elif(states_compatible(left, left_to_couple)):            
                 # if they are compatible, permute right_to_couple appropriately for proper H element
                 right_to_couple_permute = np.copy(right_to_couple)
                 if not (left_to_couple == left).all():
                     right_to_couple_permute[1:] = permute_compatible(left[1:],left_to_couple[1:],right_to_couple[1:])
                     
                 # Now first check, if the matrix element is nonzero. This is the case, if all the spins but one match up.
                 if (right[1:]==right_to_couple_permute[1:]).sum() != nspins-1:
                     continue
                 deg = degeneracy_outer_invariant_optimized(left[1:], right[1:], right_to_couple_permute[1:])
                 # check if photon number in right state increases or decreases and
                 # if all but one spin agree, and that the spin that does not agree is down in right and up in right_to_couple
                 if (right[0] - right_to_couple[0]) == 1 and sum(right[1:])-sum(right_to_couple[1:]) == 1: # need matrix element of a*sigmap
                     new_entry_func('H_g', count_out, 1j*deg * np.sqrt(right[0]))
                 elif right[0] - right_to_couple[0] == -1 and sum(right[1:])-sum(right_to_couple[1:]) == -1: # need matrix element of adag*sigmam
                     new_entry_func('H_g', count_out, 1j*deg * np.sqrt(right[0]+1))
    
            
            

        return L0_line

    @staticmethod
    def calculate_L1_line(args_tuple):
        """ Calculate L1 part of element count_in in block nu_element """
        
        #indices,count_in, nu_element = args_tuple
        nu_element, count_in = args_tuple
        global elements_block, new_entry, sparse_constructor_dic, coupled_photon_block_nu
        
        # get element, of which we want the time derivative
        current_element = elements_block[nu_element][count_in]
        current_blocksize = len(elements_block[nu_element])
    
        left = current_element[0:nspins+1] # left state, first index is photon number, rest is spin states
        right = current_element[nspins+1:2*nspins+2] # right state
        photon_tuple = (left[0], right[0])
            
        
        # Now get L1 part -> coupling from nu_element to nu_element+1
        # loop through all matrix elements in the next block we want to couple to
        next_element_block = elements_block[nu_element+1]
        next_blocksize = len(elements_block[nu_element+1])

        names = ['sigmam', 'a']
        L1_line = {name:sparse_constructor_dic((current_blocksize, next_blocksize)) for name in names}
        new_entry_func = lambda name, count_out, val: new_entry(L1_line, name, count_in, count_out, val)
        # Now get L1 part -> coupling from nu_element to nu_element+1 loop through matrix
        # elements in the next block WITH compatible photon counts (only photon number 
        # unchanged -> spin decay or photon number decreased by one -> photon decay)
        coupled_counts_nu_plus = coupled_photon_block_nu[photon_tuple][1]
        for count_out in coupled_counts_nu_plus:                   
            # get "to couple" element
            element_to_couple = next_element_block[count_out]
            left_to_couple = element_to_couple[0:nspins+1]
            right_to_couple = element_to_couple[nspins+1:2*nspins+2]
            
            #---------------------------------
            # get Liouvillian elements
            #--------------------------------
            
            # L1 part from L[sigmam] -> sigmam * rho * sigmap
            # Photons must remain the same
            if (left[0] == left_to_couple[0] and right[0] == right_to_couple[0]):
                # we have to compute matrix elements of sigma^- and sigma^+. Therefore, check first if 
                # number of spin up in "right" and "right_to_couple" as well as "left" and "left_to_coupole" vary by one
                if (sum(left[1:]) - sum(left_to_couple[1:]) == 1) and (sum(right[1:]) - sum(right_to_couple[1:]) == 1):       
                    # Get the number of permutations, that contribute.                             
                    deg = degeneracy_gamma_changing_block_efficient(left[1:], right[1:], left_to_couple[1:], right_to_couple[1:])                
                    new_entry_func('sigmam', count_out, deg)
            
            # L1 part from L[a] -> a * rho* adag
            # since spins remain the same, first check if spin states match
            # if spins match, then the element can couple, because we are looping through the block nu+1. Therefore
            # the coupled-to-elements necessarily have one more excitation, which for this case is in the photon state.
            if (left[1:] == left_to_couple[1:]).all() and (right[1:]==right_to_couple[1:]).all():
                new_entry_func('a', count_out, np.sqrt((left[0]+1)*(right[0] + 1)))
    
        return L1_line
    
    def setup_L_block_basis_parallel(self, indices, progress):
       """ Calculate Liouvillian basis in block form. Parallelize the calculation
       of rows of the Liouvillian. Called when parallel=1.
       """
       num_blocks = len(indices.mapping_block)
       #multiprocessing.set_start_method('fork')
    
       updates = 0
       if progress:
           num_elements = sum([len(indices.mapping_block[nu]) for nu in range(num_blocks)])
           bar = Progress(2*num_elements - len(indices.mapping_block[num_blocks-1]), 'Calculate L basis...') # 2 updates per block, except last block (there is no L1 for last block)
       # loop through all elements in block structure
       
       
       if indices.only_numax:
           nu_min = num_blocks - 1 # only loop through last nu if only_numax=True
       else:
           nu_min = 0
       
       for nu_element in range(nu_min, num_blocks):
           current_blocksize = len(indices.mapping_block[nu_element])
           # setup the Liouvillians for the current block
           

           L0_names = ['sigmaz', 'sigmam', 'a', 'H_n', 'H_sigmaz', 'H_g']
           L0_new = {name:self.sparse_constructor_dic((current_blocksize, current_blocksize))
                     for name in L0_names}
           
           arglist = []
           global nspins, elements_block, sparse_constructor_dic, new_entry, coupled_photon_block_nu
           nspins  = indices.nspins
           elements_block = indices.elements_block
           coupled_photon_block_nu = indices.coupled_photon_block[nu_element]
           sparse_constructor_dic = self.sparse_constructor_dic
           new_entry = self.new_entry
        #nu_element, count_in = args_tuple
           for count_in in range(current_blocksize):
               arglist.append((nu_element, count_in))
           # print(f'Block {nu_element}/{num_blocks-1}: {len(arglist)} args')
           with Pool(processes=self.num_cpus) as pool:
               #print('Number of processes:', pool._processes)
               for L0_data in pool.imap(self.calculate_L0_line, arglist):
                   for name in L0_names:
                       L0_new[name]['data'].extend(L0_data[name]['data'])
                       L0_new[name]['coords'][0].extend(L0_data[name]['coords'][0])
                       L0_new[name]['coords'][1].extend(L0_data[name]['coords'][1])
                   if progress:
                       bar.update()
                       updates +=1
           for name in L0_names:
               Lnew = L0_new[name]
               data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
               self.L0_basis[name].append(sp.coo_matrix((data, coords), shape=shape).tocsr())
           
           if nu_element < num_blocks -1:
               next_blocksize = len(indices.mapping_block[nu_element+1])
               L1_names = ['sigmam', 'a']
               L1_new = {name:self.sparse_constructor_dic((current_blocksize, next_blocksize))
                         for name in L1_names}
               with Pool(processes=self.num_cpus) as pool:

                   #print('Number of processes:', pool._processes)
                   for L1_data in pool.imap(self.calculate_L1_line, arglist):
                       for name in L1_names:
                           L1_new[name]['data'].extend(L1_data[name]['data'])
                           L1_new[name]['coords'][0].extend(L1_data[name]['coords'][0])
                           L1_new[name]['coords'][1].extend(L1_data[name]['coords'][1])
                       if progress:
                           bar.update()
                           updates+=1
               for name in L1_names:
                   Lnew = L1_new[name]
                   data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
                   self.L1_basis[name].append(sp.coo_matrix((data, coords), shape=shape).tocsr())
           # else:
           #     if progress:
           #         bar.update(2*num_elements-1)
       # print(updates, num_elements*2 - len(indices.mapping_block[-1]))

           # Loop through all elements in the same block
           # for count_in in range(current_blocksize):
           #     L0_line = self.calculate_L0_line(indices, count_in, nu_element)
               
           #     for name in L0_new:
           #         L0_new[name][count_in,:] = L0_line[name]
               
           #     if nu_element < num_blocks -1:
           #         L1_line = self.calculate_L1_line(indices, count_in, nu_element)
           #         for name in L1_new:
           #             L1_new[name][count_in,:] = L1_line[name]
            
           # append new blocks to the basis
           # for name in self.L0_basis:
           #      self.L0_basis[name].append(sp.csr_matrix(L0_new[name]))
           
           # if nu_element < num_blocks-1: 
           #     for name in self.L1_basis:
           #         self.L1_basis[name].append(sp.csr_matrix(L1_new[name]))
            
    
    
    @staticmethod
    def L0_nu_task(nu_element):
        current_blocksize = len(elements_block[nu_element])
        # setup the Liouvillians for the current block
        
        L0_names = ['sigmaz', 'sigmam', 'a', 'H_n', 'H_sigmaz', 'H_g']
        L0_new ={name:sparse_constructor_dic((current_blocksize, current_blocksize)) for name in L0_names}

        global coupled_photon_block_nu
        coupled_photon_block_nu = coupled_photon_block[nu_element]

        for count_in in range(current_blocksize):
            L0_line = BlockL.calculate_L0_line((nu_element, count_in))
            
            for name in L0_names:
                L0_new[name]['data'].extend(L0_line[name]['data'])
                L0_new[name]['coords'][0].extend(L0_line[name]['coords'][0])
                L0_new[name]['coords'][1].extend(L0_line[name]['coords'][1])
            
        return L0_new

                
    @staticmethod     
    def L1_nu_task(nu_element):
        current_blocksize = len(elements_block[nu_element])
        next_blocksize = len(elements_block[nu_element+1])

        global coupled_photon_block_nu
        coupled_photon_block_nu = coupled_photon_block[nu_element]

        L1_names = ['sigmam', 'a']
        L1_new ={name:sparse_constructor_dic((current_blocksize, next_blocksize)) for name in L1_names}
        
        for count_in in range(current_blocksize):
            L1_line = BlockL.calculate_L1_line((nu_element, count_in))
            for name in L1_names:
                L1_new[name]['data'].extend(L1_line[name]['data'])
                L1_new[name]['coords'][0].extend(L1_line[name]['coords'][0])
                L1_new[name]['coords'][1].extend(L1_line[name]['coords'][1])
        
        return L1_new

    def setup_L_block_basis_parallel2(self, indices, progress):
       """ Calculate Liouvillian basis in block form. Parallelized the calculation
       of each block"""
       num_blocks = len(indices.mapping_block)
       global nspins, elements_block, coupled_photon_block
       nspins  = indices.nspins
       elements_block = indices.elements_block
       coupled_photon_block = indices.coupled_photon_block
       
       # loop through all elements in block structure
       arglist = [nu for nu in range(num_blocks)]
       
       with Pool(processes=self.num_cpus) as pool:
           L0s = pool.map(self.L0_nu_task, arglist)
       
       arglist = arglist[:-1] # not nu_max 
       with Pool(processes=self.num_cpus) as pool:
           L1s = pool.map(self.L1_nu_task, arglist)
           
           
       L0_names = ['sigmaz', 'sigmam', 'a', 'H_n', 'H_sigmaz', 'H_g']   
       L1_names = ['a', 'sigmam']        
       
       for L0_new in L0s:
           for name in L0_names:
               Lnew = L0_new[name]
               data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
               self.L0_basis[name].append(sp.coo_matrix((data, coords), shape=shape).tocsr())
               
       for L1_new in L1s:
           for name in L1_names:
               Lnew = L1_new[name]
               data, coords, shape = Lnew['data'], Lnew['coords'], Lnew['shape']
               self.L1_basis[name].append(sp.coo_matrix((data, coords), shape=shape).tocsr())

       print('done')
       
       

class Models(BlockL):
    """ This class contains information about the exact model at hand and
    calculates the Liouvillian from the basis elements from BlockL.
    
    Demanding weak U(1) symmetry and no gain, the most general model of N spins
    interacting with a common photon mode is described by the Master equation
        
        d/dt rho = -i[H,rho] + kappa*L[a] + sum_k{ gamma*L[sigmam_k] + gamma_phi * L[sigmaz_k] }
        
    with Hamiltonian H = wc*adag*a + w0/2*sum_k{ sigmaz_k } + g*sum_k{adag*sigmam_k + a*sigmap_k }
    
    where the light-matter coupling g is assumed real.
    
    """
    def __init__(self,wc,w0,g, kappa, gamma_phi, gamma, indices, parallel=0,progress=False, debug=False, save=True, num_cpus=None, liouv_path=None, verbose=True):
        """
        

        Parameters
        ----------
        wc : float
            Cavity frequency
        w0 : float
            Two-level system level splitting
        g : float
            Light-matter coupling
        kappa : float
            Cavity loss rate
        gamma_phi : float
            Molecular dephasing rate
        gamma : float
            Molecular loss rate
        indices : setup.Indices
            Indices object
        parallel : int, optional
            Liouvillian construction:
                0: serial, 1: parallel. The default is 0.
        num_cpus : int, optional
            Number of CPUs. The default is None.
        debug : bool, optional
            If True, do not load existing file, always calculate Liouvillian from scratch. The default is False.
        save : bool, optional
            If True, save Liouvillian for later use. The default is True.
        progress : bool, optional
            Progress bar. The default is False.
        liouv_path : string, optional
            Save path for Liouvillian files. The default is None.
        verbose : bool, optional
            Print status messages. The default is True.

        Returns
        -------
        None.

        """
        
        # specify rates according to what part of Hamiltonian or collapse operators
        # they scale
        
        if indices.only_numax == True:
            if kappa != 0 or gamma != 0 :
                print(r'WARNING: only_numax is set true, but kappa = {kappa}, gamma = {gamma}. Set kappa = 0, gamma = 0 for physical results.'.format(kappa=kappa,gamma=gamma))
        
        self.rates = {'H_n': wc,
                      'H_sigmaz': w0,
                      'H_g': g,
                      'a': kappa,
                      'sigmaz': gamma_phi,
                      'sigmam': gamma}
        self.w0 = w0
        self.wc = wc
        self.g = g
        self.kappa = kappa
        self.gamma = gamma
        self.gamma_phi = gamma_phi
        self.indices = indices
        self.L0 = []
        self.L1 = []
        self.verbose = verbose
        super().__init__(indices=indices, parallel=parallel,num_cpus=num_cpus, debug=debug, save=save, progress=progress,liouv_path=liouv_path, verbose=verbose)
    
    def setup_L_Tavis_Cummings(self, progress=False, save_path=None):
        t0 = time()
        if self.indices.only_numax:
            if self.verbose:
                print('Calculating Liouvillian for TC model from basis (only nu_max) ...', flush =True)
            if progress:
                progress = False
                if self.verbose:
                    print('Disabled progress bar (only one step)')
        else:
            if self.verbose:
                print('Calculating Liouvillian for TC model from basis ...', flush =True)
        
        names0 = ['H_sigmaz', 'H_n', 'H_g','a', 'sigmam', 'sigmaz']
        names1 = ['sigmam' , 'a']
        
        
        self.L0 = []
        self.L1 = []
        
        num_blocks = len(self.indices.mapping_block)
        
        if progress: # progress bar
            loops = 2*num_blocks-1
            bar = Progress(loops,'Liouvillian: ')
            
        # Adapt loop if only_numax is true
        if self.indices.only_numax:
            nu_min = num_blocks - 1
        else:
            nu_min = 0
        
        for nu in range(nu_min, num_blocks):
            current_blocksize = len(self.indices.mapping_block[nu])
            #L0_scale = sp.csr_matrix(np.zeros((current_blocksize, current_blocksize), dtype=complex))
            L0_scale = sp.csr_matrix((current_blocksize, current_blocksize), dtype=complex)
            for name in names0:
                if self.indices.only_numax:
                    L0_scale = L0_scale + self.rates[name] * self.L0_basis[name][0] # in Liouvillian basis, if only_numax is true, only one matrix is contained (i.e. different than mapping_block, which has nu_max-1 unfilled matrices, such that the old indices still work)
                else:
                    L0_scale = L0_scale + self.rates[name] * self.L0_basis[name][nu]

            self.L0.append( L0_scale)
            
            if progress:
                bar.update()
            
            if nu < num_blocks -1:
                next_blocksize = len(self.indices.mapping_block[nu+1])
                #L1_scale = sp.csr_matrix(np.zeros((current_blocksize, next_blocksize), dtype=complex))
                L1_scale = sp.csr_matrix((current_blocksize, next_blocksize), dtype=complex)
                
                for name in names1:
                    L1_scale = L1_scale + self.rates[name] * self.L1_basis[name][nu]
                self.L1.append(L1_scale)   
                
                if progress:
                    bar.update()
 
        elapsed = time()-t0
        if self.verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        if save_path is not None:
            with open(save_path, 'wb') as handle:
                pickle.dump(self, handle)
            if self.verbose:
                print(f'Wrote full model to {save_path}.')
            
            
            
    # def setup_L_generic(self,rates,  progress=False, save_path=None):
    #     """ Calculate generic Liouvillian. IGNORES RATES GIVEN IN CONSTRUCTOR OF Models CLASS
    #     Parameters:
    #         rates (dic) : dictionary that contains the dissipation rates.
    #                      Form:
    #                           rates = {'H_n': wc,
    #                                    'H_sigmaz': w0,
    #                                    'H_g': g,
    #                                    'a': kappa,
    #                                    'sigmaz': gamma_phi,
    #                                    'sigmam': gamma,
    #                                    'sigmam_collective' : sigmam_collective,
    #                                    'ad_sigmam': ,
    #                                    'a_sigmap': }
    # """
    #     if self.verbose:
    #         print('Given rates:', rates)
    
    #     t0 = time()
    #     if self.indices.only_numax:
    #         if self.verbose:
    #             print('Calculating Liouvillian from basis (only nu_max) ...', flush =True)
    #         if progress:
    #             progress = False
    #             if self.verbose:
    #                 print('Disabled progress bar (only one step)')
    #     else:
    #         if self.verbose:
    #             print('Calculating Liouvillian from basis ...', flush =True)
        
    #     names0 = ['H_sigmaz', 'H_n', 'H_g','a', 'sigmam', 'sigmaz','ad_sigmam','a_sigmap',\
    #               'sigmam_collective']
    #     names1 = ['sigmam' , 'a', 'sigmam_collective']
        
        
    #     self.L0 = []
    #     self.L1 = []
        
    #     num_blocks = len(self.indices.mapping_block)
        
    #     if progress: # progress bar
    #         loops = 2*num_blocks-1
    #         bar = Progress(loops,'Liouvillian: ')
            
    #     # Adapt loop if only_numax is true
    #     if self.indices.only_numax:
    #         nu_min = num_blocks - 1
    #     else:
    #         nu_min = 0
        
    #     for nu in range(nu_min, num_blocks):
    #         current_blocksize = len(self.indices.mapping_block[nu])
    #         #L0_scale = sp.csr_matrix(np.zeros((current_blocksize, current_blocksize), dtype=complex))
    #         L0_scale = sp.csr_matrix((current_blocksize, current_blocksize), dtype=complex)
    #         for name in names0:
    #             if name in rates: # the key must be contained in rates dic. otherwise assume 0 value
    #                 if self.indices.only_numax:
    #                     L0_scale = L0_scale + rates[name] * self.L0_basis[name][0] # in Liouvillian basis, if only_numax is true, only one matrix is contained (i.e. different than mapping_block, which has nu_max-1 unfilled matrices, such that the old indices still work)
    #                 else:
    #                     L0_scale = L0_scale + rates[name] * self.L0_basis[name][nu]

    #         self.L0.append( L0_scale)
            
    #         if progress:
    #             bar.update()
            
    #         if nu < num_blocks -1:
    #             next_blocksize = len(self.indices.mapping_block[nu+1])
    #             #L1_scale = sp.csr_matrix(np.zeros((current_blocksize, next_blocksize), dtype=complex))
    #             L1_scale = sp.csr_matrix((current_blocksize, next_blocksize), dtype=complex)
                
    #             for name in names1:
    #                 if name in rates:
    #                     L1_scale = L1_scale + rates[name] * self.L1_basis[name][nu]
    #             self.L1.append(L1_scale)   
                
    #             if progress:
    #                 bar.update()
 
    #     elapsed = time()-t0
    #     if self.verbose:
    #         print(f'Complete {elapsed:.0f}s', flush=True)
    #     if save_path is not None:
    #         with open(save_path, 'wb') as handle:
    #             pickle.dump(self, handle)
    #         if self.verbose:
    #             print(f'Wrote full model to {save_path}.')
            

    @classmethod
    def load(cls, filepath):
        """Load a previously saved model from .pkl file"""
        with open(filepath, 'rb') as handle:
            obj = pickle.load(handle)
        # check it is actually a valid Model object...
        print(f'Loaded {type(cls)} object from {filepath}')
        return obj

class Rho:
    """ Functionality related to density matrix:
        Initial state
        Reduced density matrix
        Calculation of expectation values
    """
        
    def __init__(self, rho_p, rho_s, indices, max_nrs=1, scale_rho=1.0, verbose=True):
        """
        

        Parameters
        ----------
        rho_p : scipy.sparse.csr_matrix
            Density matrix of photons
        rho_s : scipy.sparse.csr_matrix
            2x2 density matrix of a single spin
        indices : setup.Indices
            Indices object
        max_nrs : int, optional
            Maximum number of spins contained in the reduced density matrix. The default is 1.
        scale_rho : float, optional
            Scaling factor of the density matrix. Only for numerical stability. The default is 1.0.
        verbose : bool, optional
            Print status messages. The default is True.

        Returns
        -------
        None.

        """
    
        assert type(max_nrs) == int, "Argument 'max_nrs' must be int"
        assert max_nrs >= 0, "Argument 'max_nrs' must be non-negative"
        assert indices.nspins >= max_nrs, "Number of spins in reduced density matrix "\
                "(max_nrs) cannot exceed total number of spins ({indices.nspins})"
        
        self.max_nrs = max_nrs # maximum number of spins in reduced density matrix
        self.indices = indices
        self.convert_rho_block_dic = {}
        self.initial= []
        self.scale_rho = scale_rho
        
        # debugging variables
        self.initial_full = []
        self.initial_reference = []
        
        
        # setup initial state
        t0 = time()
        if verbose:
            print('Set up initial density matrix...')
        self.initial=self.setup_initial_efficient(rho_p, rho_s)
        elapsed= time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        
        if self.scale_rho != 1:
            t0 = time()
            if verbose:
                print(f'Scaling initial density matrix by {self.scale_rho} ...')
            for nu in range(len(self.initial)):
                self.initial[nu] = self.initial[nu] * self.scale_rho
        elapsed= time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
        
        
        # setup reduced density matrix
        t0 = time()
        if verbose:
            print('Set up mappings to reduced density matrices at...')
        for nrs in range(max_nrs+1):
            if verbose:
                print(f'nrs = {nrs}...')
            self.setup_convert_rho_block_nrs(nrs)
        elapsed= time()-t0
        if verbose:
            print(f'Complete {elapsed:.0f}s', flush=True)
         
    
    def setup_initial_efficient(self, rho_p, rho_s):
        """
        Setup initial state in block representation, without first calculating
        the initial state in the compressed form
        
        Comparison: N=20 takes 185s in compressed form. Efficient version basically instant.

         Parameters
         ----------
         rho_p : sparse array
             Density matrix of initial photons state
         rho_s : sparse array
             density matrix of initial spin state

         Returns
         -------
         None

        """

        indices = self.indices
        num_blocks = len(indices.mapping_block)
        rho_vec = [np.zeros(len(i), dtype=complex) for i in indices.mapping_block] # initializing initial state

        
        # Check for superfluoresence initial condition, i.e. zero photons and all spins up. 
        # This is very easily initialized by all blocks zero, instead of the first entry of the last block
        if np.isclose(rho_p[0,0],1) and np.isclose(rho_s[0,0],1):
            rho_vec[num_blocks-1][0] = 1
            return rho_vec
        # Check for complementary condition of nu_max photons and all spins down 
        if np.isclose(rho_p[-1,-1],1) and np.isclose(rho_s[-1,-1],1):
            rho_vec[num_blocks-1][-1] = 1
            return rho_vec
            
        
        # Next, check if photon density matrix is trivial, i.e. has a one somewhere on the diagonal.
        # This is the most common use case for us, where there are exactly zero photons in the cavity mode.
        # rho_s can be general
        phot_diag = rho_p.diagonal()
        p = np.where(phot_diag == 1)[0]
        if len(p) == 1: # if there is a diagonal element of value 1 
            num_phot = p[0] # number of photons in the cavity mode -> only states with that number of photons contribute to intial state

            # Now need to loop through all elements.
            for nu in range(num_blocks):
                for element_count in range(len(indices.mapping_block[nu])):
                    element = indices.elements_block[nu][element_count]
                    p_left = element[0] # photons in left state
                    p_right = element[indices.nspins+1] # photons in right state
                    
                    # Where left and right photons match num_phot, the 
                    # entry can be nonzero, depending on the spin state
                    if p_left == p_right and p_left == num_phot:
                        rho_vec[nu][element_count] = 1                        
                        s_left = element[1:indices.nspins+1] # left spin state
                        s_right = element[indices.nspins+2:] # right spin state
                        
                        # loop through spin states and multiply all contributions
                        for count_s in range(len(s_left)):
                            rho_vec[nu][element_count] *= rho_s[s_left[count_s], s_right[count_s]]
            return rho_vec
            
        
        
        # now completely general initial state; probably never needed, because we almost always
        # set the photon number to zero in the beginning.
        # Loop through all elements
        for nu in range(num_blocks):
            for element_count in range(len(indices.mapping_block[nu])):
                element = indices.elements_block[nu][element_count]
                p_left = element[0] # photons in left state
                p_right = element[indices.nspins+1] # photons in right state
                
                rho_vec[nu][element_count] = rho_p[p_left, p_right]

                if rho_vec[nu][element_count] != 0:
                    s_left = element[1:indices.nspins+1] # left spin state
                    s_right = element[indices.nspins+2:] # right spin state
                    
                    # loop through spin states and multiply all contributions
                    for count_s in range(len(s_left)):
                        rho_vec[nu][element_count] *= rho_s[s_left[count_s], s_right[count_s]]
        return rho_vec
                        
                        

    def setup_convert_rho_block_nrs(self, nrs):
        """Setup conversion matrix from supercompressed vector to vector form of
        reduced density matrix of the photon plus nrs (possibly 0) spins

        N.B. Takes advantage of counts of how many spins are different between
        ket (left) and bra (right) for a state with a given excitation nu and
        block_index, as stored in indices.different_block_inv

        Fills in self.convert_rho_block_dic with an entry with key nrs
        """
        indices = self.indices
        nspins = indices.nspins
        num_elements = [len(block) for block in indices.mapping_block]

        # initialise empty matrices for each block of the supercompressed form
        convert_rho_block = [
                sp.lil_matrix(((indices.ldim_p*indices.ldim_s**nrs)**2, num), dtype=float)
                for num in num_elements
                ]

        self.convert_rho_block_dic[nrs] = convert_rho_block # modified in place 

        # for reduced density matrix involving nrs spins, must consider elements
        # with a most nrs differences between ket (left) and bra (right) spin
        # states
        for num_diff in range(nrs+1):
            # difference_block_inv has a tuple (nu, block_index) describing all the
            # block elements and indices with a given num_diff differences
            block_element_tuples = indices.difference_block_inv[num_diff]
            for nu, block_index in block_element_tuples:
                # to calculation contribution to rdm, we need element
                element = indices.elements_block[nu][block_index]
                count_p1, left, count_p2, right = \
                        element[0], element[1:nspins+1], element[nspins+1], element[nspins+2:] 
                diff_arg = (left != right).nonzero()[0] # location of spins with different ket, bra states 
                same = np.delete(left, diff_arg) # spins with states same in ket and bra
                self.add_all_for_block_element(nrs, nu, block_index,
                                               count_p1, left[diff_arg],
                                               count_p2, right[diff_arg],
                                               same)
        convert_rho_block = [block.tocsr() for block in convert_rho_block]
    
    def add_all_for_block_element(self, nrs, nu, block_index,
                                  count_p1, left, count_p2, right, same, 
                                  s_start=0):
        """Populate all entries in conversion matrix at nu with reduced density matrix indices
        associated with permutations of spin values |left> and <right| and column index
        block_index, according to the number of permutations of spin values in 'same'.

        nrs is the number of spins in the target reduced density matrix ('number reduced spins').
        """
        if len(left) == nrs:
            # add contributions from 'same' to rdm at |left><right|
            self.add_to_convert_rho_block_dic(nrs, nu, block_index,
                                              count_p1, left,
                                              count_p2, right,
                                              same)
            return # end of recursion
        # current |left> too short for rdm, so move element from 'same' to |left> (and <right|)
        # iterate through all possible values of spin...
        for s in range(s_start, self.indices.ldim_s):
            s_index = next((i for i,sa in enumerate(same) if sa==s), None)
            # ...but only act on the spins that are actually in 'same'
            if s_index is None:
                continue
            # extract spin value from same, append to left and right
            tmp_same = np.delete(same, s_index)
            tmp_left = np.append(left, s)
            tmp_right = np.append(right, s)
            # repeat until |left> and <right| are correct length for rdm
            self.add_all_for_block_element(nrs, nu, block_index,
                                           count_p1, tmp_left,
                                           count_p2, tmp_right,
                                           tmp_same, s_start=s) # can skip up to s in next function call
    
    def add_to_convert_rho_block_dic(self, nrs, nu, block_index,
                                     count_p1, diff_left,
                                     count_p2, diff_right,
                                     same):
        """Calculate contribution to reduced density matrix element with spin state
        |diff_left><diff_right| (photon |count_p1><count_p1|) according to free
        permutations of 'same' """
        convert_rho_block = self.convert_rho_block_dic[nrs]
        # number of permutations of spins in same, each of which contributes one unit under trace 
        combinations = _multinominal(np.bincount(same))
        # get all vectorised reduced density matrix indices for element
        s_indices = np.arange(nrs)
        rdm_indices = []
        for perm_indices in permutations(s_indices):
            # spins in rdm are still identical, so need to populate elements are
            # all rdm indices associated with (distinct) permutations of the nrs spin
            # N.B. duplication occurs here, use more_itertools.distinct_permutations() to
            # avoid; only costly for large nrs (?)
            index_list = list(perm_indices)
            rdm_indices.append(self.get_rdm_index(count_p1, diff_left[index_list],
                                                  count_p2, diff_right[index_list]))
        for rdm_index in rdm_indices:
            convert_rho_block[nu][rdm_index, block_index] = combinations
            
    def get_rdm_index(self, count_p1, left, count_p2, right):
        """Calculate index in vectorised reduced density matrix 
        for element |count_p1><count_p2|(X)|left><right|

        This index is according to column-stacking convention used by qutip; see e.g.

        A=qutip.Qobj(numpy.arange(4).reshape((2, 2))
        print(qutip.operator_to_vector(A))

        I can't remember writing this magic - pfw
        """
        ket = np.concatenate(([count_p1], left))
        bra = np.concatenate(([count_p2], right))
        row = 0
        column = 0
        nrs = len(ket)-1
        for i in range(nrs+1):
            j = nrs-i
            row += ket[j] * self.indices.ldim_s**i
            column += bra[j] * self.indices.ldim_s**i
        return row + column * self.indices.ldim_p * self.indices.ldim_s**nrs
    
    
    

if __name__ == '__main__':
    # Testing purposes
    
    # same parameters as in Peter Kirton's code.
    ntls =2#number 2LS
    nphot = ntls+1
    w0 = 1.0
    wc = 0.65
    Omega = 0.4
    g = Omega / np.sqrt(ntls)
    kappa = 0.011
    gamma = 0.02
    gamma_phi = 0.03
    indi = Indices(ntls)













